"""Send daily or weekly notification emails."""
from __future__ import annotations

from flask import render_template
from flask_mail import Message
from flask_sqlalchemy import SQLAlchemy

from labster.di import injector
from labster.domain2.model.demande import Demande
from labster.domain2.model.notification import Notification
from labster.domain2.model.profile import Profile
from labster.extensions import mail
from labster.rpc.queries.demandes_tables import mes_taches

db = injector.get(SQLAlchemy)


#
# Notifications (activity streams)
#
def send_notification_to(user: Profile) -> None:
    pending_notifications = get_pending_notifications(user)
    if not pending_notifications:
        return

    html = make_notification_email(user, pending_notifications)

    recipient = user.email
    subject = "Vos notifications Lab&Co"
    send_email(recipient, subject, html)

    for notification in pending_notifications:
        notification.sent = True


def make_notification_email(user: Profile, notifications: list[Notification]) -> str:
    demandes_set = {notification.demande for notification in notifications}
    demandes = sorted(demandes_set, key=lambda x: x.created_at, reverse=True)

    def get_notifications(demande: Demande) -> list[Notification]:
        return [n for n in notifications if n.demande == demande]

    ctx = {
        "demandes": demandes,
        "get_notifications": get_notifications,
    }
    return render_template("notifications/email/notifications.html", **ctx)


#
# Recaps
#
def send_recap_to(user: Profile) -> bool:
    html = make_recap_email(user)
    if not html:
        return False

    recipient = user.email
    subject = "Le point hebdomadaire sur vos demandes sur Lab&Co"
    send_email(recipient, subject, html)
    return True


def make_recap_email(user: Profile) -> str | None:
    """Returns None if nothing to send."""
    notifications = get_pending_notifications(user)

    demandes_set = set(mes_taches(user))
    has_tasks = bool(demandes_set)

    for notif in notifications:
        demandes_set.add(notif.demande)

    demandes = sorted(demandes_set, key=lambda x: x.created_at, reverse=True)

    if not demandes:
        return None

    def get_notifications(demande: Demande) -> list[Notification]:
        return [n for n in notifications if n.demande == demande]

    def get_cta(demande: Demande) -> str:
        workflow = demande.get_workflow(user)
        state = demande.get_state(user)
        if user not in state.task_owners(workflow):
            return ""

        state_id = state.id
        mapping = {
            "EN_VALIDATION": "Valider la demande",
            "EN_EDITION": "Compléter et soumettre la demande",
            "EN_VERIFICATION": "Vérifier la demande",
            "EN_INSTRUCTION": "Traiter ou rejeter la demande",
        }
        return mapping[state_id]

    ctx = {
        "demandes": demandes,
        "has_tasks": has_tasks,
        "get_notifications": get_notifications,
        "get_cta": get_cta,
    }
    return render_template("notifications/email/recap-new.html", **ctx)


#
# Util
#
def get_pending_notifications(user: Profile) -> list[Notification]:
    notifications = (
        Notification.query.filter(Notification.user == user)
        .filter(Notification.sent == False)
        .order_by(Notification.created_at)
        .all()
    )
    return notifications


def send_email(recipient, subject, html):
    msg = Message(subject, recipients=[recipient], html=html)
    mail.send(msg)



# from __future__ import annotations
#
# from functools import singledispatch
#
# from labster.domain.models.demandes import Demande, DemandeAvenantConvention, \
#     DemandeConvention, DemandeRH
# from labster.domain.services.calculs_couts import get_ctx_for_demande
# from labster.extensions import db
#
# from .model import StatsLine
#
#
# def sync_all():
#     StatsLine.query.delete()
#
#     demandes = Demande.query.all()
#     for demande in demandes:
#         line = sync(demande)
#         if line:
#             db.session.add(line)
#
#     db.session.commit()
#
#
# def sync(demande):
#     if demande.wf_state == "EN_EDITION":
#         keep = False
#         for state in demande.wf_history:
#             transition = state.get("transition")
#             if transition in (
#                 "REQUERIR_MODIFICATION_DIR",
#                 "REQUERIR_MODIFICATION_DGRTT",
#             ):
#                 keep = True
#         if not keep:
#             return
#
#     data = {
#         "demande_id": demande.id,
#         "type_demande": demande.type,
#         "wf_state": demande.wf_state,
#         "date_soumission": demande.date_soumission,
#         "date_finalisation": demande.date_finalisation,
#     }
#
#     if demande.porteur:
#         data["porteur_id"] = demande.porteur_id
#     else:
#         data["porteur_id"] = None
#
#     structure = demande.structure
#     if structure:
#         data["structure_id"] = structure.id
#         data["equipe_id"] = get_id(structure.equipe)
#         data["departement_id"] = get_id(structure.departement)
#         data["labo_id"] = get_id(structure.laboratoire)
#         data["ufr_id"] = get_id(structure.ufr)
#         data["pole_id"] = get_id(structure.pole)
#
#     update_data = sync_specific(demande)
#     # if update_data is None:
#     #     return None
#
#     data.update(update_data)
#     line = StatsLine(**data)
#     return line
#
#
# @singledispatch
# def sync_specific(demande):
#     data = {}
#     return data
#
#
# @sync_specific.register(DemandeConvention)
# def _(demande):
#     data = {
#         "montant": pick_integer(demande, "montant_financement"),
#         "recrutements_prev": pick_integer(demande, "nombre_recrutements"),
#         "duree": pick_integer(demande, "duree_previsionnelle"),
#         "financeur": getattr(demande, "type_financeur", ""),
#     }
#     return data
#
#
# @sync_specific.register(DemandeAvenantConvention)
# def _1(demande):
#     data = {"financeur": getattr(demande, "type_financeur", "")}
#     return data
#
#
# @sync_specific.register(DemandeRH)
# def _2(demande):
#     # print(demande.id)
#     data = {
#         "duree": demande.duree_mois or 0,
#         "type_recrutement": (
#             getattr(demande, "nature_du_recrutement", "")
#             + " | "
#             + getattr(demande, "type_de_demande", "")
#         ),
#     }
#
#     ctx = get_ctx_for_demande(demande)
#     data["salaire_brut_mensuel"] = ctx["salaire_brut"]
#     data["cout_total_mensuel"] = ctx["cout_total"]
#
#     return data
#
#
# # # TODO
# # @sync_specific.register(DemandePiLogiciel)
# # def _3(demande):
# #     data = {}
# #     return data
# #
# #
# # # TODO
# # @sync_specific.register(DemandePiInvention)
# # def _4(demande):
# #     data = {}
# #     return data
# #
# #
# # # TODO
# # @sync_specific.register(DemandeAutre)
# # def _5(demande):
# #     data = {}
# #     return data
#
#
# def pick_integer(demande, field_name):
#     try:
#         value = int(getattr(demande, field_name).replace(" ", ""))
#         if value > 2147483647:
#             value = 2147483647
#     except (ValueError, AttributeError):
#         # print(getattr(demande, field_name, None))
#         # traceback.print_exc()
#         value = 0
#     return value
#
#
# def get_id(obj):
#     if obj:
#         return obj.id
#     else:
#         return None
from __future__ import annotations



from __future__ import annotations

from flask import Blueprint

blueprint = Blueprint("bi", __name__, template_folder="templates", url_prefix="/bi")
route = blueprint.route


@blueprint.record
def configure(state):
    from . import views



from __future__ import annotations

from copy import deepcopy
from typing import Any

from flask import json
from flask_sqlalchemy import SQLAlchemy

from labster.di import injector
from labster.domain2.model.structure import StructureRepository
from labster.domain2.model.type_structure import DE, EQ
from labster.domain2.services.constants import get_constants
from labster.domain2.services.roles import Role, RoleService
from labster.types import JSONDict

from .fields import Field, FieldSet

structure_repo = injector.get(StructureRepository)
role_service = injector.get(RoleService)
db = injector.get(SQLAlchemy)


class Form:
    fieldsets: list[FieldSet] = []
    conditions: str = ""
    mode: str = "edit"
    name: str = ""
    label: str = ""

    def __init__(
        self, model=None, mode="edit", laboratoire=None, porteur=None, gestionnaire=None
    ):
        self.fieldsets = deepcopy(self.__class__.fieldsets)
        self.mode = mode

        self.post_init()

        if porteur:
            self.set_porteur(porteur)

        elif gestionnaire:
            self.set_porteurs_possibles(gestionnaire)

        elif laboratoire:
            self.set_membres_du_labo(laboratoire)

        if model is None:
            self.model = self.empty_model()
        elif isinstance(model, dict):
            self.model = model
        else:
            raise RuntimeError("Unimplemented")

    def post_init(self):
        pass

    def update_choices(self):
        """Override in subclasses."""

        self.set_structures_concernees_choices()
        self.set_contributeurs_choices()

    def set_porteurs_possibles(self, gestionnaire):
        return

        # TODO
        # roles = gestionnaire.get_roles(RoleType.GDL)
        # membres = set()
        # for role in roles:
        #     org = role.context
        #     porteurs_org = [x for x in org.get_membres() if x.has_role("porteur")]
        #     membres.update(porteurs_org)
        #
        # porteurs_du_labo = sorted(membres, key=lambda x: x.nom)
        # field = self.get_field("porteur")
        # if field:
        #     field.choices = [(m.uid, f"{m.nom} {m.prenom}") for m in porteurs_du_labo]

    def set_membres_du_labo(self, labo):
        membres_du_labo = sorted(labo.membres, key=lambda x: x.nom)
        field = self.get_field("porteur")
        if field:
            field.choices = [(m.uid, f"{m.nom} {m.prenom}") for m in membres_du_labo]

    def set_porteur(self, porteur):
        field = self.get_field("porteur")
        if field:
            field.choices = [(porteur.uid, f"{porteur.nom} {porteur.prenom}")]
            field.editable = False

    def to_dict(self) -> JSONDict:
        return {
            "name": self.name,
            "label": self.label,
            "fieldsets": [fs.to_dict() for fs in self.fieldsets],
            "fields": {f.name: f.to_dict() for f in self.fields},
            "csrf_token": "",
            "mode": self.mode,
            "constants": get_constants(),
            "conditions": self.conditions,
        }

    def to_json(self) -> str:
        return json.dumps(
            self.to_dict(), sort_keys=True, indent=4, separators=(",", ": ")
        )

    @property
    def fields(self) -> list[Field]:
        fields = []
        for fs in self.fieldsets:
            for f in fs.fields:
                if isinstance(f, Field):
                    fields.append(f)
        return fields

    def get_field(self, name: str) -> Any:
        for f in self.fields:
            if f.name == name:
                return f
        return None

    def empty_model(self) -> JSONDict:
        model: JSONDict = {}
        for field in self.fields:
            model[field.name] = ""
        return model

    def fix_model(self, model: JSONDict) -> JSONDict:
        model = deepcopy(model)
        for field in self.fields:
            if field.name not in model:
                model[field.name] = ""
        return model

    #
    #
    #
    def set_structures_concernees_choices(self):
        field = self.get_field("structures_concernees")
        if not field:
            return

        structures = structure_repo.get_all()
        choices = [
            {"value": s.id, "label": f"{s.type}: {s.nom} ({s.sigle})"}
            for s in structures
            if s.type not in {DE, EQ}
        ]
        choices.sort(key=lambda t: t["label"])
        field.choices = choices

    def set_contributeurs_choices(self):
        field = self.get_field("contributeurs")
        if not field:
            return

        profiles = role_service.get_users_with_role(Role.PORTEUR, "*")
        profiles = [p for p in profiles if p.active]
        profiles.sort(key=lambda p: (p.nom, p.prenom))

        choices = [{"value": p.id, "label": f"{p.nom}, {p.prenom}"} for p in profiles]
        choices.sort(key=lambda t: t["label"])
        field.choices = choices



from __future__ import annotations

import os
import socket
from datetime import timedelta
from typing import Any

from abilian.web.action import Endpoint
from flask.cli import load_dotenv

load_dotenv()


HOST_MAP = {
    "orophin": "TIRETDEV",
    "tareg": "PREPROD",
    "zirak": "PRODUCTION",
}

DEFAULT_SRC = [
    "'self'",
    "https://cdn.ckeditor.com/",
    "https://unpkg.com",
    "https://maxcdn.bootstrapcdn.com",
]

CONTENT_SECURITY_POLICY = {
    "default-src": DEFAULT_SRC,
    "script-src": DEFAULT_SRC
    + [
        "'unsafe-inline'",
        "'unsafe-eval'",
        "https://sentry.io/",
        "https://browser.sentry-cdn.com/",
    ],
    "connect-src": DEFAULT_SRC
    + [
        "https://sentry.io/",
    ],
    "style-src": DEFAULT_SRC + ["'unsafe-inline'"],
    "img-src": DEFAULT_SRC + ["'unsafe-inline'", "data:"],
    "font-src": DEFAULT_SRC + ["'unsafe-inline'", "data:"],
}

DEV_CONTENT_SECURITY_POLICY = {"default-src": ["'*'"], "script-src": ["'*'"]}


class DefaultConfig:
    NAME = "DEFAULT"
    PRODUCTION = False

    # FIXME later
    WTF_CSRF_ENABLED = False

    CONTENT_SECURITY_POLICY = CONTENT_SECURITY_POLICY

    # Flask config
    PERMANENT_SESSION_LIFETIME = timedelta(days=1)
    SECRET_KEY = "tototiti"

    # abilian-core config
    SITE_NAME = "Lab&Co Sorbonne Université"
    PRIVATE_SITE = False
    MAIL_ASCII_ATTACHMENTS = True
    ANTIVIRUS_CHECK_REQUIRED = True
    LOGO_URL = Endpoint("static", filename="img/logo_carre_32px.jpg")
    FAVICO_URL = Endpoint("static", filename="img/logo-su-square.png")

    MAIL_SENDER = "noreply@sorbonne-universite.fr"
    MAIL_SUPPRESS_SEND = True

    CELERYBEAT_SCHEDULE: dict[Any, Any] = {}
    #     # Executes every day at 6 A.M
    #     "add-every-monday-morning": {
    #         "task": "tasks.add",
    #         "schedule": crontab(hour=6, minute=0),
    #     }
    # }

    # Babel config
    BABEL_ACCEPT_LANGUAGES = ("fr",)
    BABEL_DEFAULT_TIMEZONE = "Europe/Paris"

    # Celery config
    REDIS_URI = "redis://localhost:6379/1"
    BROKER_URL = "redis://localhost:6379/0"
    CELERYD_PREFETCH_MULTIPLIER = 1
    CELERY_ALWAYS_EAGER = True
    CELERY_EAGER_PROPAGATES_EXCEPTIONS = True
    CELERY_RESULT_BACKEND = "redis://localhost:6379/0"

    # Persistence
    SQLALCHEMY_DATABASE_URI = "postgresql+psycopg2://localhost/labster"

    MAIL_SERVER = "localhost"

    # APISPEC_SWAGGER_URL = "/labster-api"
    # APISPEC_SWAGGER_UI_URL = "/labster-api-ui"
    OPENAPI_VERSION = "3.0.2"

    # Auth
    CAS_SERVER = "https://auth.sorbonne-universite.fr/cas/"
    ALLOW_BACKDOOR = False

    RESTHEART_URL = "http://localhost:18080/db"
    RESTHEART_AUTH = ("admin", "DfgV0UYgwdMM")

    # Logs
    LOG_DB = "data/log.db"


class DevConfig(DefaultConfig):
    NAME = "DEV"
    SERVER_NAME = "localhost:5000"

    # Relax security cookies settings (which are more stringent by default)
    SESSION_COOKIE_SECURE = False
    REMEMBER_COOKIE_SECURE = False
    SESSION_COOKIE_HTTPONLY = False
    REMEMBER_COOKIE_HTTPONLY = False
    CONTENT_SECURITY_POLICY = DEV_CONTENT_SECURITY_POLICY

    # Debug
    DEBUG = True
    DEBUG_TB_ENABLED = True
    DEBUG_TB_INTERCEPT_REDIRECTS = False
    DEBUG_TB_PROFILER_ENABLED = False
    DEBUG_TB_PANELS = [
        "flask_debugtoolbar.panels.versions.VersionDebugPanel",
        "flask_debugtoolbar.panels.timer.TimerDebugPanel",
        "flask_debugtoolbar.panels.headers.HeaderDebugPanel",
        "flask_debugtoolbar.panels.request_vars.RequestVarsDebugPanel",
        "flask_debugtoolbar.panels.template.TemplateDebugPanel",
        # 'flask_debugtoolbar.panels.sqlalchemy.SQLAlchemyDebugPanel',
        "flask_debugtoolbar.panels.logger.LoggingPanel",
        "flask_debugtoolbar.panels.profiler.ProfilerDebugPanel",
        "abilian.services.indexing.debug_toolbar.IndexedTermsDebugPanel",
        # 'flask_debugtoolbar_lineprofilerpanel.panels.LineProfilerPanel',
    ]
    TEMPLATE_DEBUG = True
    ALLOW_BACKDOOR = True

    # SQLALCHEMY_ECHO = True


class DemoConfig(DefaultConfig):
    NAME = "DEMO"

    PRODUCTION = True
    PREFERRED_URL_SCHEME = "https"
    SERVER_NAME = "labster-dev.demo.abilian.com"

    MAIL_SUPPRESS_SEND = False
    TEST_EMAIL_ADRESS = "labandco-test@abilian.net"


class PreprodConfig(DefaultConfig):
    NAME = "PREPROD"

    PRODUCTION = True
    PREFERRED_URL_SCHEME = "https"
    SERVER_NAME = "labandco-pre.sorbonne-universite.fr"

    MAIL_SUPPRESS_SEND = False
    TEST_EMAIL_ADRESS = "labandco-test@abilian.net"


class TiretDevConfig(DefaultConfig):
    NAME = "TIRETDEV"

    PRODUCTION = True
    PREFERRED_URL_SCHEME = "https"
    SERVER_NAME = "labandco-dev.sorbonne-universite.fr"

    MAIL_SUPPRESS_SEND = False
    TEST_EMAIL_ADRESS = "labandco-test@abilian.net"


class ProdConfig(DefaultConfig):
    NAME = "PRODUCTION"
    SERVER_NAME = "labandco.sorbonne-universite.fr"

    PRODUCTION = True
    PREFERRED_URL_SCHEME = "https"
    MAIL_SUPPRESS_SEND = False

    # Debug email
    # EMAIL_CC = ['X']


CONFIG_MAP = {
    "DEV": DevConfig,
    "DEMO": DemoConfig,
    "TIRETDEV": TiretDevConfig,
    "PREPROD": PreprodConfig,
    "PRODUCTION": ProdConfig,
}


def get_config():
    env = os.environ
    name = env.get("APP_NAME", "DEV")

    # Hack
    hostname = socket.gethostname()
    if hostname in HOST_MAP:
        name = HOST_MAP[hostname]

    config_class = CONFIG_MAP.get(name)
    if not config_class:
        raise RuntimeError(f"Unknown application name: {name}")
    config = config_class()

    for k, v in os.environ.items():
        if k.startswith("APP_"):
            config_key = k[len("APP_") :]
            config_value = v
            setattr(config, config_key, config_value)

    # Override SQLALCHEMY_DATABASE_URI w/ env variables if needed
    pg_password = env.get("PG_PASSWORD", "")
    database_url = env.get("DATABASE_URL", "")
    sqlalchemy_database_uri = env.get("SQLALCHEMY_DATABASE_URI", "")

    if database_url:
        config.SQLALCHEMY_DATABASE_URI = database_url
    elif sqlalchemy_database_uri:
        config.SQLALCHEMY_DATABASE_URI = sqlalchemy_database_uri
    elif pg_password:
        db_uri = f"postgresql+psycopg2://labster:{pg_password}@localhost/labster"
        config.SQLALCHEMY_DATABASE_URI = db_uri

    return config



from __future__ import annotations

import json
from os.path import dirname
from typing import Any

import structlog
from dotted.collection import DottedCollection, DottedDict

from labster.domain.models.config import Config
from labster.extensions import db

"""

- adding new constants: add their type in TYPES and their values in
constants.json. get_constant will check for the new constants.

- upgrading the value of existing constants: use _upgrade_if_needed and the version number.

"""


logger = structlog.get_logger()

TYPES = {
    "api_key": "str",
    "single_user": "str",
    "convention.COUT_ENVIRONNEMENT_PERSONNEL_HEBERGE": "int",
    "convention.COUT_HORAIRE_STAGE": "float",
    "convention.DUREE_AMORTISSEMENT": "list[list[str,int]]",
    "convention.REMUNERATION": "list[list[str,int]]",
    "convention.TAUX_CHARGE_PATRONALE": "float",
    "convention.TAUX_ENVIRONNEMENT_PERSONNEL_NON_PERMANENT": "float",
    "convention.TAUX_ENVIRONNEMENT_PERSONNEL_PERMANENT": "float",
    "convention.TAUX_PARTICIPATION_COUTS_INDUITS": "float",
    "convention.TAUX_PROVISION_RISQUE": "float",
    "faq_categories": "list[str]",
    "message_dgrtt": "HTML",
    "nom_bureaux_dgrtt.MSAR": "str",
    "nom_bureaux_dgrtt.PIJ": "str",
    "nom_bureaux_dgrtt.ETI": "str",
    "nom_bureaux_dgrtt.AIPI": "str",
    "nom_bureaux_dgrtt.CC": "str",
    "nom_bureaux_dgrtt.CFE": "str",
    "nom_bureaux_dgrtt.GF": "str",
    "nom_bureaux_dgrtt.EU": "str",
    "nom_bureaux_dgrtt.PI2": "str",
    "nom_bureaux_dgrtt.CP": "str",
    "nom_bureaux_dgrtt.REF": "str",
    "nom_bureaux_dgrtt.DIR": "str",
    "nom_bureaux_dgrtt.CT": "str",
    "nom_bureaux_dgrtt.ETT": "str",
    "point_indice": "float",
    "recrutement.charges_moins_12_mois": "float",
    "recrutement.charges_plus_12_mois": "float",
    "recrutement.grades": "list[str]",
    "recrutement.salaire_brut_mensuel_indicatif.Chercheur": "str",
    "recrutement.salaire_brut_mensuel_indicatif.Post-doctorant": "list[list[str,int]]",
    "recrutement.salaire_brut_mensuel_indicatif.IGR": "list[list[str,int]]",
    "recrutement.salaire_brut_mensuel_indicatif.IGE": "list[list[str,int]]",
    "recrutement.salaire_brut_mensuel_indicatif.ASI": "list[list[str,int]]",
    "recrutement.salaire_brut_mensuel_indicatif.technicien-sup": "list[list[str,int]]",
    "recrutement.salaire_brut_mensuel_indicatif.technicien-normale": "list[list[str,int]]",
    "recrutement.salaire_brut_mensuel_indicatif.adjoint": "list[list[str,int]]",
    "recrutement.principes": "HTML",
    "recrutement.provision_risque_charge_employeur": "float",
    "recrutement.salaire_min_doctorant": "float",
    "recrutement.couts_charges.PU/DR C EX Confirmé": "str",
    "recrutement.couts_charges.PU/DR C EX": "str",
    "recrutement.couts_charges.PU/DR C1 Confirmé": "str",
    "recrutement.couts_charges.PU/DR C1": "str",
    "recrutement.couts_charges.PU/DR C2 Confirmé": "str",
    "recrutement.couts_charges.PU/DR C2": "str",
    "recrutement.couts_charges.MC/CR HC Confirmé": "str",
    "recrutement.couts_charges.MC/CR HC": "str",
    "recrutement.couts_charges.MC/CR CN Confirmé": "str",
    "recrutement.couts_charges.MC/CR CN": "str",
    "recrutement.couts_charges.IR HC Confirmé": "str",
    "recrutement.couts_charges.IR HC": "str",
    "recrutement.couts_charges.IR C1 Confirmé": "str",
    "recrutement.couts_charges.IR C1": "str",
    "recrutement.couts_charges.IR C2 Confirmé": "str",
    "recrutement.couts_charges.IR C2": "str",
    "recrutement.couts_charges.IE HC confirmé": "str",
    "recrutement.couts_charges.IE HC": "str",
    "recrutement.couts_charges.IE C1 Confirmé": "str",
    "recrutement.couts_charges.IE C1": "str",
    "recrutement.couts_charges.IE C2 Confirmé": "str",
    "recrutement.couts_charges.IE C2": "str",
    "recrutement.couts_charges.ASI Confirmé": "str",
    "recrutement.couts_charges.ASI": "str",
    "recrutement.couts_charges.TR CE Confirmé": "str",
    "recrutement.couts_charges.TR CE": "str",
    "recrutement.couts_charges.TR CS Confirmé": "str",
    "recrutement.couts_charges.TR CS": "str",
    "recrutement.couts_charges.TR CN Confirmé": "str",
    "recrutement.couts_charges.TR CN": "str",
    "recrutement.transport": "float",
    "demande_recrutement.pj": "HTML",
    "demande_recrutement.bareme": "HTML",
    "pi_invention.pieces_a_joindre": "HTML",
    "pi_invention.conditions": "HTML",
    "pi_logiciel.conditions": "HTML",
}

VALID_TYPES = [
    "int",
    "float",
    "bool",
    "str",
    "HTML",
    "list[str]",
    "list[list[str,int]]",
]

for t in TYPES.values():
    assert t in VALID_TYPES, t


def get_message_dgrtt():
    return get_constant("message_dgrtt", "OK")


def get_faq_categories():
    categories = get_constant("faq_categories", [])
    return [(x, x) for x in categories]


def _get_constants() -> dict[str, Any]:
    """Get constants from config or local json with updating system.

    Pick new constants that are defined in TYPES but are not saved in
    the config yet from the json.

    Upgrades constants that already exist, given the version number.
    """
    from labster.domain.models.config import Config

    config = Config.query.first()

    if config is None:
        config = Config()
        db.session.add(config)

    initial_constants = get_initial_constants()

    # upgrade
    _upgrade_if_needed(config, initial_constants)

    constants = config.data
    dotted_constants = DottedCollection.factory(constants)
    json_dotted_constants = DottedCollection.factory(initial_constants)

    if dotted_constants:
        for key in TYPES.keys():
            # do not write "key not in .keys()", it doesn't return "dotted keys".
            if dotted_constants.get(key, _MARKER) is _MARKER:
                dotted_constants[key] = json_dotted_constants.get(key)

        constants = dotted_constants.to_python()

    return constants


def _upgrade_if_needed(config: Config, initial_constants: dict[str, Any]):
    """ "data migration" tool.

    - constants: the ones from the config, to update.
    - initial_constants: from the json file, that may have additions.
    """
    needs_commit = False
    version = initial_constants.get("version")

    if not version:
        return

    constants = config.data or {}

    # Check the state we want rather than a version number.
    # if version == 0.2 and constants.get('version', 0) < 0.2:
    if "recrutement" not in constants:
        constants["recrutement"] = {}
        needs_commit = True

    if "grades" not in constants["recrutement"]:
        constants["recrutement"]["grades"] = []
        needs_commit = True

    if (
        "recrutement" in constants
        and constants["recrutement"]["grades"]
        and constants["recrutement"]["grades"][0] == "IR"
    ):
        constants["recrutement"]["grades"] = initial_constants["recrutement"]["grades"]
        constants["version"] = version
        needs_commit = True

    if version == 0.3:
        # add salaire_brut_mensuel_indicatif.doctorat-plus-3
        # Maybe a new DB doesn't have salaire_brut_mensuel_indicatif yet.
        if not constants["recrutement"].get("salaire_brut_mensuel_indicatif"):
            constants["recrutement"][
                "salaire_brut_mensuel_indicatif"
            ] = initial_constants["recrutement"]["salaire_brut_mensuel_indicatif"]
        else:
            post_doct_names = [
                it[0]
                for it in constants["recrutement"]["salaire_brut_mensuel_indicatif"][
                    "Post-doctorant"
                ]
            ]
            if "doctorat-plus-3" not in post_doct_names:
                constants["recrutement"]["salaire_brut_mensuel_indicatif"][
                    "Post-doctorant"
                ].append(
                    initial_constants["recrutement"]["salaire_brut_mensuel_indicatif"][
                        "Post-doctorant"
                    ][-1]
                )

        # logger.info("--- constants updated for 0.3 - doctorat-plus-3")
        constants["version"] = 0.3
        needs_commit = True

    if needs_commit:
        config.data = constants
        db.session.commit()


def get_initial_constants() -> dict[str, Any]:
    filename = dirname(__file__) + "/constants.json"
    return json.load(open(filename, "rb"))


def get_constants() -> dict[str, Any]:
    constants = _get_constants()
    return update_constants(constants)


_MARKER = object()


def get_constant(path: str, default: object = _MARKER) -> Any:
    """Return this constant's value from a dotted path.

    Raises a KeyError if the path is illegal.

    - path: a dotted key (str) (example: "convention.REMUNERATION")
    - returns: the value or a default one (of the good type, if specified).
    """
    if path not in TYPES:
        raise KeyError(path)

    constants = _get_constants()
    dotted_constants = DottedCollection.factory(constants)

    try:
        value = dotted_constants[path]
        if isinstance(value, DottedCollection):
            return dotted_constants[path].to_python()
        else:
            return value
    except KeyError:
        pass

    # If a default value is supplied, return it
    if default != _MARKER:
        return default

    # Otherwise, use a default depending on the type
    type_ = TYPES[path]
    return default_value(type_)


def default_value(type_: str) -> Any:
    if type_ in ["str", "HTML"]:
        return ""
    elif type_ == "int":
        return 0
    elif type_ == "float":
        return 0.0
    elif type_ == "bool":
        return False
    elif type_.startswith("list"):
        return []
    else:
        raise RuntimeError(f"Unknown type: {type_}")


def save_constants(constants):
    from labster.domain.models.config import Config

    updated_constants = update_constants(constants)

    config = Config.query.first()
    if config is None:
        config = Config(data=updated_constants)
    else:
        config.data = updated_constants
    return config


def update_constants(constants: dict[str, Any]) -> dict[str, Any]:
    if "types" in constants:
        del constants["types"]
    dotted_constants = DottedCollection.factory(constants)

    updated_constants = DottedDict()
    for key, type_ in TYPES.items():
        try:
            # do not write "key in .keys()", it doesn't check for dotted keys.
            if dotted_constants.get(key, _MARKER) is not _MARKER:
                value = dotted_constants[key]
            elif key in TYPES.keys():
                value = get_constant(key)
            else:
                value = default_value(type_)
        except KeyError:
            value = default_value(type_)

        try:
            value = coerce(value, type_)
        except TypeError:
            msg = "Wrong type for key: {}, value{} "
            raise TypeError(msg.format(key, value))
        check_type(key, value, type_)
        updated_constants[key] = value

    return updated_constants.to_python()


def check_type(key: str, value: Any, type_: str) -> None:
    def ensure_types(types):
        val = value
        if isinstance(value, DottedCollection):
            val = value.to_python()
        if not isinstance(val, types):
            msg = f"Wrong type for key: {key}. Expected: {type_}, got: {type(val)}"
            raise TypeError(msg)

    if type_ == "int":
        ensure_types(int)
    elif type_ == "float":
        ensure_types((float, int))
    elif type_ == "str":
        ensure_types(str)
    elif type_ == "bool":
        ensure_types(bool)
    elif type_.startswith("list"):
        ensure_types(list)
        # TODO: more specific cases


def coerce(value: Any, type_: str) -> Any:
    if type_ == "int":
        return int(value)
    elif type_ == "float":
        return float(value)
    elif type_ == "str":
        return str(value)
    elif type_ == "bool":
        return bool(value)
    else:
        return value



"""Champs communs à plusieurs formulaires."""
from __future__ import annotations

from labster.newforms.base import FieldSet, Select2Field
from labster.newforms.base.fields import MultipleSelect2Field

laboratoire = FieldSet(  #
    "laboratoire",
    "Structure demandeuse",
    [
        Select2Field(
            "laboratoire",
            "Nom de la structure demandeuse",
            choices=[],
            required=True,
            editable=True,
        )
    ],
)

porteur = FieldSet(  #
    "porteur",
    "Porteur de la demande",
    [Select2Field("porteur", "Porteur de la demande", choices=[], required=True)],
)

structures_concernees = FieldSet(  #
    "structures_concernees",
    "Structures concernées",
    [
        MultipleSelect2Field(
            "structures_concernees", "Structures concernées", choices=[]
        )
    ],
)

contributeurs = FieldSet(  #
    "contributeurs",
    "Contributeurs autorisés à prendre la main sur la demande",
    [MultipleSelect2Field("contributeurs", "Contributeurs", choices=[])],
)



# from __future__ import annotations
#
# from typing import Dict, List, Optional, Set, Tuple
#
# from injector import inject
#
# from labster.domain2.model.profile import Profile, ProfileId, ProfileRepository
# from labster.domain2.model.structure import Structure, StructureId, \
#     StructureRepository
# from labster.domain2.services.contacts import ContactService, ContactType
#
#
# class InmemoryContactService(ContactService):
#     edges: Set[Tuple[str, ContactType, str]]
#
#     @inject
#     def __init__(
#         self, profile_repo: ProfileRepository, structure_repo: StructureRepository,
#     ):
#         self.profile_repo = profile_repo
#         self.structure_repo = structure_repo
#         self.edges = set()
#
#     def get_state(self):
#         return [self.edges]
#
#     def set_state(self, state):
#         self.edges = state[0]
#
#     def is_empty(self):
#         return len(self.edges) == 0
#
#     def clear(self):
#         self.edges.clear()
#
#     def set_contact(
#         self, structure: Structure, contact_type: ContactType, user: Profile
#     ):
#         assert isinstance(structure, Structure)
#         assert isinstance(contact_type, ContactType)
#         assert isinstance(user, Profile)
#
#         if self.get_contact(structure, contact_type) == user:
#             return
#
#         self.delete_contact(structure, contact_type)
#         edge = (structure.id, contact_type, user.id)
#         self.edges.add(edge)
#
#     def delete_contact(self, structure: Structure, contact_type: ContactType):
#         assert isinstance(structure, Structure)
#         assert isinstance(contact_type, ContactType)
#
#         for edge in list(self.edges):
#             if (edge[0], edge[1]) == (structure.id, contact_type):
#                 self.edges.remove(edge)
#
#     def get_contact(
#         self, structure: Structure, contact_type: ContactType
#     ) -> Optional[Profile]:
#         for edge in self.edges:
#             if (structure.id, contact_type) != edge[0:2]:
#                 continue
#             user_id = edge[2]
#             user = self.profile_repo.get_by_id(ProfileId(user_id))
#             return user
#         return None
#
#     def get_mapping(self) -> Dict[Structure, Dict[ContactType, Profile]]:
#         result = {}
#         contact_types: List[ContactType] = list(ContactType)
#
#         structure_ids = {edge[0] for edge in self.edges}
#
#         structures = [
#             self.structure_repo.get_by_id(StructureId(structure_id))
#             for structure_id in structure_ids
#         ]
#         structures2: List[Structure] = [s for s in structures if s]
#         structures2.sort(key=lambda x: x.name)
#
#         for structure in structures2:
#             d: Dict[ContactType, Optional[Profile]] = {
#                 key: None for key in contact_types
#             }
#             for edge in self.edges:
#                 if edge[0] == structure.id:
#                     contact_type = edge[1]
#                     user = self.profile_repo.get_by_id(ProfileId(edge[2]))
#                     if user:
#                         d[contact_type] = user
#             result[structure] = d
#
#         return result
from __future__ import annotations



"""Modèle de case management adapté au projet Labster."""
from __future__ import annotations

from typing import TYPE_CHECKING

import structlog

from labster.lib.workflow import State, Workflow

if TYPE_CHECKING:
    from labster.domain2.model.profile import Profile

logger = structlog.get_logger()


class EnEdition(State):
    label = "En édition"
    next_action = "Edition à finaliser et à soumettre"

    def on_enter(self, workflow):
        case = workflow.case
        case.active = True
        case.editable = True

    def task_owners(self, workflow):
        case = workflow.case
        return {u for u in [case.gestionnaire, case.porteur] if u}


class EnValidation(State):
    label = "En cours de validation"
    label_short = "En validation"
    next_action = "Demande à considérer pour validation"

    def task_owners(self, workflow: Workflow) -> set[Profile]:
        demande = workflow.case
        assert demande

        structures_signataires = demande.structures_signataires()
        signatures = workflow.get_value("signatures") or []

        result = set()
        for structure in structures_signataires:
            if structure.id not in signatures:
                result.update(structure.responsables)

        return result


class EnVerification(State):
    label = "Recevabilité en cours de vérification"
    label_short = "En vérification"
    next_action = "Recevabilité à confirmer"

    def on_enter(self, workflow):
        workflow.set_value("ar_envoye", False)

    def task_owners(self, workflow):
        case = workflow.case
        if case.contact_labco:
            return {case.contact_labco}
        else:
            return set()


class EnInstruction(State):
    label = "En cours d'instruction par la DR&I"
    label_short = "En instruction"
    next_action = "Instruction à mener et finaliser"

    def task_owners(self, workflow):
        case = workflow.case
        if case.contact_labco:
            return {case.contact_labco}
        else:
            return set()


# Etats finaux
class Traitee(State):
    label = "Traitée par la DR&I"
    label_short = "Traitée"
    is_final = True


class Rejetee(State):
    label = "Rejetée par la DR&I"
    label_short = "Rejetée"
    is_final = True


class Abandonnee(State):
    label = "Abandonnée par le porteur"
    label_short = "Abandonnée"
    is_final = True


EN_EDITION = EnEdition()
EN_VALIDATION = EnValidation()
EN_VERIFICATION = EnVerification()
EN_INSTRUCTION = EnInstruction()
TRAITEE = Traitee()
REJETEE = Rejetee()
ABANDONNEE = Abandonnee()

ACTIVE_STATES: list[State] = [
    EN_EDITION,
    EN_VALIDATION,
    EN_VERIFICATION,
    EN_INSTRUCTION,
]
INACTIVE_STATES: list[State] = [TRAITEE, REJETEE, ABANDONNEE]
ALL_STATES: list[State] = ACTIVE_STATES + INACTIVE_STATES



from __future__ import annotations

from flask_sqlalchemy import SQLAlchemy
from jsonrpcserver import method

from labster.di import injector
from labster.domain2.model.profile import Profile, ProfileRepository
from labster.domain2.model.structure import StructureRepository
from labster.domain2.services.roles import Role, RoleService
from labster.rpc.cache import cache

structure_repo = injector.get(StructureRepository)
profile_repo = injector.get(ProfileRepository)
role_service = injector.get(RoleService)
db = injector.get(SQLAlchemy)


@method
def update_membres_rattaches(structure_id: str, values: list[dict]):
    structure = structure_repo.get_by_id(structure_id)
    assert structure

    membres: list[Profile] = role_service.get_users_with_given_role(
        Role.MEMBRE_RATTACHE, structure
    )

    current_membre_ids = {m.id for m in membres}
    updated_membre_ids = {m["id"] for m in values}

    membres_to_add = updated_membre_ids.difference(current_membre_ids)
    for user_id in membres_to_add:
        user = profile_repo.get_by_id(user_id)
        role_service.grant_role(user, Role.MEMBRE_RATTACHE, structure)

    membres_to_remove = current_membre_ids.difference(updated_membre_ids)
    for user_id in membres_to_remove:
        user = profile_repo.get_by_id(user_id)
        role_service.ungrant_role(user, Role.MEMBRE_RATTACHE, structure)

    db.session.commit()

    cache.evict("users")
    cache.evict("structures")



from __future__ import annotations

from flask_sqlalchemy import SQLAlchemy
from jsonrpcserver import method

from labster.di import injector
from labster.domain2.model.demande import Demande
from labster.security import get_current_profile


@method
def wf_transition(demande_id, action, data=None):
    data = data or {}
    user = get_current_profile()
    db = injector.get(SQLAlchemy)

    demande = db.session.query(Demande).get(demande_id)
    workflow = demande.get_workflow(user)

    try:
        transition = workflow.get_transition_by_id(action)
    except IndexError:
        msg = (
            "Action impossible. Quelqu'un a probablement effectué une action "
            "sur la demande en parallèle avec vous.",
            "danger",
        )
        return msg

    workflow.execute_transition(transition, data=data)
    db.session.commit()
    msg = (
        f"Votre action '{transition.label}' sur la demande '{demande.nom}' a bien été prise en compte.",
        "success",
    )
    return msg


# @method
# def wf_transition(demande_id, action):
#     user = get_current_profile()
#     db = injector.get(SQLAlchemy)
#
#     form = transition.get_form(workflow)
#     for field in form:
#         field.form = form
#     is_form_valid = form.validate()
#
#     if __action == "confirm":
#         if is_form_valid:
#             data = {}
#             for field in form:
#                 data[field.id] = field.data
#             workflow.execute_transition(transition, data=data)
#             db.session.commit()
#             flash(
#                 "Votre action '{}' sur la demande '{}' a bien été prise en compte.".format(
#                     transition.label, demande.nom
#                 )
#             )
#             return redirect(url_for(demande, _external=True))
#         else:
#             flash(
#                 "Merci de bien vouloir corriger ou compléter les informations "
#                 "ci-dessous.",
#                 "danger",
#             )
#             flash(f"{form.errors}")
#
#     title = "Confirmer l'action"
#     breadcrumbs = [{"name": "Demandes", "url": url_for(".demandes")}, {"name": title}]
#
#     ctx = {
#         "title": title,
#         "breadcrumbs": breadcrumbs,
#         "form": form,
#         "transition": transition,
#         "demande": demande,
#     }
#     return render_template("wf/confirm.html", **ctx)



from __future__ import annotations

import re
from io import BytesIO
from typing import Any

import structlog
from attr import attrs
from flask_sqlalchemy import SQLAlchemy
from ldif import LDIFParser
from tqdm import tqdm

from labster.di import injector
from labster.domain2.model.profile import Profile, ProfileRepository
from labster.domain2.model.structure import Structure, StructureRepository
from labster.ldap.constants import ADMINS_DN, PRESIDENCE_DN, SU_DN, \
    get_parent_dn

logger = structlog.get_logger()

profile_repo = injector.get(ProfileRepository)
structure_repo = injector.get(StructureRepository)
db = injector.get(SQLAlchemy)


def parse_ldif_file(ldif_file: str) -> list[tuple[str, dict[str, Any]]]:
    logger.info(f"### Parsing LDIF file {ldif_file}")

    orig_ldif_fd = open(ldif_file, "rb")
    ldif_fd = BytesIO()
    for line in orig_ldif_fd.readlines():
        if line.startswith(b"# search result"):
            break
        ldif_fd.write(line)

    ldif_fd.seek(0)

    parser = LDIFParser(ldif_fd)
    return list(parser.parse())


@attrs(auto_attribs=True)
class LdifRecord:
    raw: dict[str, list[str]]

    def __getattr__(self, name):
        return self.raw.get(name, [""])[0]

    @property
    def uid(self) -> str | None:
        if "uid" not in self.raw:
            return None
        return self.raw["uid"][0]

    @property
    def affectation(self) -> str | None:
        structure_affectaction = self._get_structure_d_affectation()
        if not structure_affectaction:
            return None

        affectation = structure_affectaction.dn
        if affectation in ADMINS_DN:
            affectation = get_parent_dn(affectation)

        if affectation == PRESIDENCE_DN:
            affectation = SU_DN

        return affectation

    def _get_structure_d_affectation(self) -> Structure | None:
        structure_d_affectation = None

        affectation_principale = self.supannEntiteAffectationPrincipale
        if affectation_principale:
            structure_d_affectation = (
                db.session.query(Structure)
                .filter(Structure.supann_code_entite == affectation_principale)
                .first()
            )
            if structure_d_affectation:
                return structure_d_affectation

        # Old LDIF format
        affectation = self.sorbonneUniversiteEmpAffectation
        if affectation:
            structure_d_affectation = (
                db.session.query(Structure).filter(Structure.dn == affectation).first()
            )
            if structure_d_affectation:
                return structure_d_affectation

        return None

    @property
    def fonctions(self):
        return self.raw.get("eduPersonAffiliation", [])

    @property
    def address(self):
        adresse = self.raw.get("postalAddress", [""])[0]
        adresse = adresse.replace("$", "\n")
        adresse = re.sub("\n\n+", "\n\n", adresse)
        adresse = adresse.strip()
        return adresse


def update_users_from_records(records: list[tuple[str, dict[str, list[str]]]]):
    profiles = profile_repo.get_all()
    old_profile_uids = {
        profile.uid for profile in profiles if profile.uid and profile.active
    }
    count0 = len(old_profile_uids)
    print(f"old total: {count0:d}")
    logger.info(f"old total: {count0:d}")

    new_profile_uids = set()
    for _dn, r in records:
        if "uid" not in r:
            continue
        uid = r["uid"][0]
        new_profile_uids.add(uid)

    deleted_uids = old_profile_uids.difference(new_profile_uids)
    deactivate_users(deleted_uids)

    uids_to_profiles = {p.uid: p for p in profiles}

    print("Updating profiles from LDIF dump")
    for _dn, r in tqdm(records, disable=None):
        record = LdifRecord(r)
        if not record.uid:
            continue
        uid = record.uid
        if not uid:
            continue
        if uid in uids_to_profiles:
            profile = uids_to_profiles[uid]
        else:
            profile = Profile(uid=uid)
            profile_repo.put(profile)

        update_profile_from_record(profile, record)


def deactivate_users(deleted_uids):
    logger.info("To be deactivated:", deleted_uids=deleted_uids)

    for uid in tqdm(deleted_uids, disable=None):
        user = profile_repo.get_by_uid(uid)
        user.deactivate()


def update_profile_from_record(profile: Profile, record: LdifRecord):
    assert profile

    profile.nom = record.sn
    profile.prenom = record.givenName
    profile.uid = record.uid
    profile.email = record.mail
    profile.telephone = record.telephoneNumber
    profile.adresse = record.address
    profile.login = record.supannAliasLogin
    profile.adresse = record.adresse

    affectation = record.affectation

    if not affectation:
        if profile.active:
            profile.affectation = ""
            profile.deactivate()
        return

    if not profile.active:
        profile.activate()

    if profile.affectation != affectation:
        profile.affectation = affectation

    fonctions = list(record.fonctions)
    if set(profile.fonctions) != set(fonctions):
        profile.fonctions = fonctions



from __future__ import annotations

from abilian.services.indexing.adapter import SAAdapter

from labster.domain.models.demandes import Demande
from labster.domain.models.faq import FaqEntry


def register(app):
    index_service = app.services["indexing"]
    index_service.adapters_cls.insert(0, DemandeAdapter)
    index_service.adapters_cls.insert(0, FaqAdapter)


class DemandeAdapter(SAAdapter):
    @staticmethod
    def can_adapt(obj_cls):
        return issubclass(obj_cls, Demande)

    def get_document(self, obj):
        kw = super().get_document(obj)

        indexed_values = []
        for k in ["nom", "name", "no_infolab", "no_eotp"]:
            v = getattr(obj, k)
            if v:
                indexed_values.append(v)

        for v in obj.data.values():
            if isinstance(v, str):
                indexed_values.append(v)

        allowed_roles_and_users = []

        v = obj.porteur
        if v:
            indexed_values.append(v.full_name)
            allowed_roles_and_users += [f"user:{v.id}"]

        v = obj.gestionnaire
        if v:
            indexed_values.append(v.full_name)
            allowed_roles_and_users += [f"user:{v.id}"]

        v = obj.laboratoire
        if v:
            indexed_values.append(v.nom)

        v = obj.structure
        if v:
            indexed_values.append(v.nom)
            allowed_roles_and_users += [f"org:{v.id}"]

        kw["text"] = " ".join(indexed_values)
        kw["name"] = obj.nom

        kw["allowed_roles_and_users"] = " ".join(allowed_roles_and_users)
        return kw


class FaqAdapter(SAAdapter):
    @staticmethod
    def can_adapt(obj_cls):
        return issubclass(obj_cls, FaqEntry)

    def get_document(self, obj):
        kw = super().get_document(obj)

        indexed_values = []
        for k in ["title", "body"]:
            v = getattr(obj, k)
            if v:
                indexed_values.append(v)

        kw["text"] = " ".join(indexed_values)
        kw["allowed_roles_and_users"] = "all"
        return kw



from __future__ import annotations

from enum import Enum

from labster.domain2.model.profile import Profile
from labster.domain2.model.structure import Structure


class ContactType(Enum):
    CONTACT_ENTREPRISES = "Contact Entreprises"
    CONTACT_EUROPE = "Contact Europe"
    CONTACT_CONTRATS_PUBLICS = "Contact Contrats Publics"
    CONTACT_RH = "Contact RH"
    CONTACT_PI = "Contact PI"
    JURISTE_PI = "Juriste PI"
    CONTACT_DRV = "Contact DRV"
    CDP = "CDP"


class ContactService:
    def is_empty(self):
        raise NotImplementedError

    def clear(self):
        raise NotImplementedError

    def set_contact(
        self, structure: Structure, contact_type: ContactType, user: Profile
    ):
        raise NotImplementedError

    def delete_contact(self, structure: Structure, contact_type: ContactType):
        raise NotImplementedError

    def get_contact(
        self, structure: Structure, contact_type: ContactType
    ) -> Profile | None:
        raise NotImplementedError

    def get_mapping(self) -> dict[Structure, dict[ContactType, Profile]]:
        raise NotImplementedError

    def get_mapping_for(self, structure: Structure) -> dict[ContactType, Profile]:
        raise NotImplementedError



from __future__ import annotations

from tqdm import tqdm

from labster.di import injector
from labster.domain2.model.profile import Profile, ProfileRepository
from labster.domain2.model.structure import StructureRepository
from labster.domain2.services.roles import Role, RoleService


class RolesUpdater:
    def __init__(self):
        self.profile_repo = injector.get(ProfileRepository)
        self.structure_repo = injector.get(StructureRepository)
        self.role_service = injector.get(RoleService)

    def update_roles(self, max=0):
        profiles = self.profile_repo.get_all()
        print("nbre de profils", len(profiles))
        if max:
            profiles = list(profiles)[0:max]
        for profile in tqdm(profiles, disable=None):
            if not profile.active:
                continue
            self.update_roles_for(profile)

    def update_roles_for(self, user: Profile):
        self.update_role_membre_affecte(user)
        self.role_service.update_roles(user)

        # TODO
        # self.update_role_porteur(user)

    def update_role_membre_affecte(self, user: Profile):
        roles = self.role_service.get_roles_for_user(user)
        structures_actuelles = list(roles.get(Role.MEMBRE_AFFECTE, []))
        if (
            len(structures_actuelles) == 1
            and structures_actuelles[0].dn == user.affectation
        ):
            return

        for structure in structures_actuelles:
            self.role_service.ungrant_role(user, Role.MEMBRE_AFFECTE, structure)

        structure = self.structure_repo.get_by_dn(user.affectation)
        if structure:
            self.role_service.grant_role(user, Role.MEMBRE_AFFECTE, structure)

    # def update_role_porteur(self, user: Profile):
    #     if "researcher" in user.fonctions or "faculty" in user.fonctions:
    #         self.remove_role_porteur(user)
    #         structure = self.structure_repository.get_by_dn(user.affectation)
    #         self.role_service.grant_role(user, Role.PORTEUR, structure)
    #
    #     else:
    #         self.remove_role_porteur(user)
    #
    # def remove_role_porteur(self, user: Profile):
    #     roles = self.role_service.get_roles_for_user(user)
    #     structures = roles.get(Role.PORTEUR, [])
    #     for structure in structures:
    #         self.role_service.ungrant_role(user, Role.PORTEUR, structure)



from __future__ import annotations

from werkzeug.exceptions import Forbidden

from labster.di import injector
from labster.domain2.services.roles import Role, RoleService
from labster.security import get_current_user

role_service = injector.get(RoleService)


def ensure_role(role: Role):
    user = get_current_user()
    if user.is_anonymous:
        return

    profile = user.profile
    if not profile.has_role(role):
        raise Forbidden()


def owner_sorter(owner):
    is_signataire = role_service.has_role(owner, Role.SIGNATAIRE, "*")
    return (-int(is_signataire), owner.nom, owner.prenom)



from __future__ import annotations

import os
import tempfile
from pathlib import Path

from flask import Blueprint, Flask, request, send_file
from werkzeug.exceptions import BadRequest, Forbidden, NotFound

from labster.auth import AuthContext
from labster.domain2.services.roles import Role

blueprint = Blueprint("backup", __name__, url_prefix="/backup")
route = blueprint.route

__all__ = ()

ROOT = "/home/labster"


@route("/db")
def backup_db(app: Flask, auth_context: AuthContext):
    check_admin(auth_context)

    config = app.config
    if "PG_PASSWORD" in config:
        os.putenv("PGUSER", "labster")
        os.putenv("PGPASSWORD", config["PG_PASSWORD"])

    fd, name = tempfile.mkstemp()
    status = os.system(f"pg_dump -F c labster > {name}")
    assert status == 0
    os.unlink(name)
    fp = os.fdopen(fd, "rb")
    return send_file(fp, as_attachment=True, attachment_filename="labster.dump")


@route("/file/")
def backup_file(app: Flask, auth_context: AuthContext):
    check_admin(auth_context)

    path_arg = request.args.get("path", "")
    path = Path(ROOT) / path_arg
    path = path.resolve()

    if not str(path).startswith(ROOT):
        raise BadRequest()

    if not path.exists():
        raise NotFound()

    if path.is_dir():
        files = list(path.glob("*"))
        files.sort()
        result = [f"<a href='.?path={file}'>{file}</a>" for file in files]
        return "<br>\n".join(result), {"content-type": "text/html"}

    else:
        fp = path.open("rb")
        return send_file(fp, as_attachment=True, attachment_filename=path.name)


def check_admin(auth_context: AuthContext):
    user = auth_context.current_profile
    assert user

    if not user.has_role(Role.ADMIN_CENTRAL):
        raise Forbidden()



"""Debug tooling."""
from __future__ import annotations

import sys
import time
from functools import wraps

from flask_sqlalchemy import connection_stack


def timeit(method):
    @wraps(method)
    def timed(*args, **kw):
        ctx = connection_stack.top
        ctx.sqlalchemy_queries = []

        ts = time.time()
        result = method(*args, **kw)
        te = time.time()

        # Print report
        print(78 * "-")
        if "log_time" in kw:
            name = kw.get("log_name", method.__name__.upper())
            kw["log_time"][name] = int((te - ts) * 1000)
        else:
            print(f"{method.__name__!r} {(te - ts) * 1000:2.2f} ms")
            if args:
                print(f"args: {args}")
            if kw:
                print(f"kwargs: {kw}")

        queries = ctx.sqlalchemy_queries
        sql_time = sum(query.duration for query in queries)
        print(f"SQL queries (# = {len(queries)}, time={sql_time*1000}ms):")

        # for i, query in enumerate(queries):
        #     print(i, query.statement)
        #     print(i, query.parameters)
        #     print(i, query.duration)
        #     print(i, query.context)
        # print(f"SQL queries (# = {len(queries)}, time={sql_time*1000}ms):")

        print(78 * "-")
        print()
        sys.stdout.flush()
        return result

    return timed



from __future__ import annotations

from labster.domain2.services.roles import Role
from labster.lib.menu import Menu
from labster.test.test_web import login_as_dgrtt

MAIN = {
    "label": "Menu principal",
    "entries": [
        {"label": "Accueil", "endpoint": "main.home", "icon": "home"},
        {
            "label": "Tâches",
            "endpoint": "demandes.tasks",
            "icon": "check-square",
            "requires_role": {"recherche", "dgrtt"},
        },
        {
            "label": "Demandes en cours",
            "endpoint": "demandes.demandes",
            "icon": "briefcase",
            "requires_role": {"recherche", "dgrtt"},
        },
        {
            "label": "Demandes archivées",
            "endpoint": "demandes.archives",
            "icon": "graduation-cap",
            "requires_role": {"recherche", "dgrtt"},
        },
        {
            "label": "Questions & suggestions",
            "endpoint": "faq.home",
            "icon": "question",
        },
        {
            "label": "Statistiques",
            "endpoint": "bi.home",
            "icon": "chart-line",
            "requires_role": {
                Role.ADMIN_CENTRAL,
                Role.RESPONSABLE,
                # "alc",
                # "directeur",
                # "chef de bureau",
                # "gouvernance",
                # "direction dgrtt",
            },
        },
        {
            "label": 'Calculette "devis RH"',
            "endpoint": "demandes.calculette_devis_rh",
            "icon": "calculator",
        },
        {
            "label": 'Calculette "feuille de coûts"',
            "endpoint": "demandes.calculette_feuille_cout",
            "icon": "calculator",
        },
    ],
}


def test_menu_not_empty(client, db_session):
    login_as_dgrtt(client, db_session)
    menu = Menu(MAIN)
    assert not menu.is_empty()


def test_menu_serialize(client, db_session):
    login_as_dgrtt(client, db_session)
    menu = Menu(MAIN)
    dto = menu.asdict()
    assert isinstance(dto, dict)



from __future__ import annotations

from labster.domain2.services.roles import Role
from labster.lib.menu import Menu
from labster.rbac import can_view_stats

MAIN = {
    "label": "Menu principal",
    "entries": [
        # Homes
        {"label": "Accueil", "to": "/", "icon": "home"},
        {
            "label": "Tâches",
            "to": "/tasks",
            "icon": "check-square",
        },
        {
            "label": "Mes demandes en cours",
            "to": "/demandes",
            "icon": "briefcase",
        },
        {
            "label": "Demandes archivées",
            "to": "/archives",
            "icon": "graduation-cap",
        },
        # Stuff
        {"label": "Questions & suggestions", "to": "/faq", "icon": "question"},
        {
            "label": "Statistiques",
            "to": "/bi",
            "icon": "chart-line",
            "precondition": can_view_stats,
        },
        {
            "label": 'Calculette "devis RH"',
            "to": "/calculette_rh",
            "icon": "calculator",
        },
        {
            "label": 'Calculette "feuille de coûts"',
            "endpoint": "main.calculette_feuille_cout",
            "icon": "calculator",
        },
    ],
}

ANNUAIRES = {
    "label": "Annuaires",
    "entries": [
        {"label": "Structures", "to": "/annuaire/structures", "icon": "sitemap"},
        {"label": "Utilisateurs", "to": "/annuaire/users", "icon": "user"},
        {
            "label": "Contacts Lab&Co",
            "to": "/contacts",
            "icon": "arrows-alt",
            "requires_role": {"alc"},
        },
    ],
}

ADMIN = {
    "label": "Administration",
    "entries": [
        # {"label": "Home", "icon": "home", "requires_role": {"alc"},},
        # {"label": "Tableau de bord", "icon": "chart-line", "requires_role": {"alc"},},
        {
            "label": "Gérer la FAQ",
            "icon": "question",
            "to": "/admin/faq",
            "requires_role": {Role.ADMIN_CENTRAL, Role.FAQ_EDITOR},
        },
        {
            "label": "Config",
            "icon": "sliders-h",
            "to": "/admin/constants",
            "requires_role": {"alc"},
        },
        {
            "label": "Rôles globaux",
            "icon": "lock",
            "to": "/admin/roles",
            "requires_role": {"alc"},
        },
        # {
        #     "label": "Financeurs",
        #     "icon": "euro-sign",
        #     # "endpoint": "admin2.financeurs",
        #     "requires_role": {"alc"},
        # },
    ],
}


def get_menu(user):
    menus = [Menu(MAIN), Menu(ANNUAIRES), Menu(ADMIN)]

    return [m for m in menus if not m.is_empty()]



from __future__ import annotations

from abilian.web.forms import Form, optional, required
from wtforms import BooleanField, StringField, TextAreaField


class WorkflowForm(Form):
    note = TextAreaField("Note", validators=[optional()])
    resoumission = BooleanField(
        "Redemander validation par la hiérarchie", validators=[optional()]
    )

    def __init__(self, require_note=False, ask_for_revalidation=False, *args, **kw):
        super().__init__(*args, **kw)

        if require_note:
            self["note"].validators = [required()]
        else:
            self["note"].validators = [optional()]

        if not ask_for_revalidation:
            del self["resoumission"]


class ConfirmerRecevabiliteForm(WorkflowForm):
    no_infolab = StringField("N° Infolab / référence DRI", validators=[required()])


class ConfirmerFinalisationForm(WorkflowForm):
    no_eotp = StringField("N° EOTP / référence DRI", validators=[required()])



from __future__ import annotations

from typing import TYPE_CHECKING

from attr import attrs

if TYPE_CHECKING:
    from labster.domain2.model.structure import Structure

void: frozenset[TypeStructure] = frozenset()


@attrs(hash=True, auto_attribs=True, frozen=True, str=False, repr=False)
class TypeStructure:
    name: str
    can_have_multiple_parents: bool
    reel: bool
    types_parents: frozenset[TypeStructure]

    def check(self, structure: Structure) -> bool:
        return structure.type == self

    def full_check(self, structure: Structure) -> None:
        msg = f"check failed on {structure}: {structure.type_name} != {self.name}"
        assert self.check(structure), msg

        parents = structure.parents
        for parent in parents:
            msg = (
                f"check failed on {structure}: "
                f"{parent.type_name} ∉ {self.types_parents}"
            )
            assert any(type.check(parent) for type in self.types_parents), msg

    def __str__(self):
        return self.name

    def __repr__(self) -> str:
        return f"TypeStructure({self.name})"

    def can_have_parent_of_type(self, type: TypeStructure) -> bool:
        return type in self.types_parents

    def can_have_child_of_type(self, type: TypeStructure) -> bool:
        return type.can_have_parent_of_type(self)

    @property
    def id(self):
        for k, v in globals().items():
            if v == self:
                return k


def get_type_structure(name: str) -> TypeStructure:
    for obj in globals().values():
        if not isinstance(obj, TypeStructure):
            continue
        if obj.name.lower() == name.lower():
            return obj

    raise KeyError(name)


def get_type_structure_by_id(id: str) -> TypeStructure:
    for obj in globals().values():
        if not isinstance(obj, TypeStructure):
            continue
        if obj.id.lower() == id.lower():
            return obj

    raise KeyError(id)


f = frozenset

UN = TypeStructure("Université", False, True, void)
DU = TypeStructure("Direction universitaire", False, True, f({UN}))
SU = TypeStructure("Service universitaire", False, True, f({DU}))

FA = TypeStructure("Faculté", False, True, f({UN}))
DF = TypeStructure("Direction facultaire", False, True, f({FA}))
SF = TypeStructure("Service facultaire", False, True, f({DF}))

CO = TypeStructure("Composante", False, True, f({FA}))

CA = TypeStructure("Carnot", True, False, f({CO}))
LX = TypeStructure("Labex", True, False, f({CO}))
IN = TypeStructure("Institut", True, False, f({CO}))
GRC = TypeStructure("GRC", True, False, f({CO}))
INI = TypeStructure("Initiative", True, False, f({CO}))

ED = TypeStructure("École doctorale", False, True, f({DU}))
LA = TypeStructure("Laboratoire", False, True, f({CO, CA, LX, IN, GRC}))

DE = TypeStructure("Département", False, False, f({LA}))
EQ = TypeStructure("Équipe", False, False, f({LA, DE}))

ALL_TYPES = [UN, FA, DU, CO, SU, DF, LA, SF, ED, DE, EQ, CA, LX, IN, GRC, INI]



"""Methodes JSON-RPC pour manipuler le graphe des structures.

Note: les méthodes sont préfixées par "sg_" (sg = "structures graph").

Roles / Permissions:

Admin central:

- L’admin central est le seul qui peut créer des sous-structures dans des structures autres
  que les labos
- Seul l’admin central peut créer des structures virtuelles

cf. https://trello.com/c/cRUEKsVv/
"""
from __future__ import annotations

from flask_sqlalchemy import SQLAlchemy
from jsonrpcserver import method
from werkzeug.exceptions import NotFound

from labster.auth import AuthContext
from labster.di import injector
from labster.domain2.model.profile import ProfileRepository
from labster.domain2.model.structure import Structure, StructureId, \
    StructureRepository
from labster.domain2.model.type_structure import get_type_structure_by_id
from labster.domain2.services.roles import RoleService
from labster.rbac import check_permission, check_structure_editable
from labster.rpc.cache import cache
from labster.types import JSON

structure_repo = injector.get(StructureRepository)
profile_repo = injector.get(ProfileRepository)
role_service = injector.get(RoleService)
auth_context = injector.get(AuthContext)
db = injector.get(SQLAlchemy)


@method
def sg_update_structure(id: str, model: dict[str, JSON]):
    structure = structure_repo.get_by_id(StructureId(id))
    check_structure_editable(structure)

    for k, v in model.items():
        setattr(structure, k, v)

    db.session.commit()
    cache.evict("structures")


@method
def sg_create_child_structure(id: str, model: dict[str, str]):
    parent_structure = structure_repo.get_by_id(StructureId(id))
    if not parent_structure:
        raise NotFound()

    check_permission(parent_structure, "P3")

    new_structure = Structure()
    new_structure.nom = model["nom"]
    type_structure = get_type_structure_by_id(model["type_id"])
    new_structure.type_name = type_structure.name

    parent_structure.add_child(new_structure)
    structure_repo.put(new_structure)

    db.session.commit()
    cache.evict("structures")


@method
def sg_delete_structure(id: str):
    structure = structure_repo.get_by_id(StructureId(id))
    if not structure:
        raise NotFound()

    check_permission(structure, "P3")

    structure.delete()

    db.session.commit()
    cache.evict("structures")


@method
def sg_add_edge(u_id, v_id):
    u = structure_repo.get_by_id(StructureId(u_id))
    v = structure_repo.get_by_id(StructureId(v_id))

    check_permission(u, "P3")

    u.add_child(v)

    db.session.commit()
    cache.evict("structures")


@method
def sg_delete_edge(u_id, v_id):
    u = structure_repo.get_by_id(StructureId(u_id))
    v = structure_repo.get_by_id(StructureId(v_id))

    check_permission(u, "P3")

    u.remove_child(v)

    db.session.commit()
    cache.evict("structures")



from __future__ import annotations

from abilian.web.forms import Form, RichTextWidget, Select2Field, required, \
    strip
from wtforms import StringField


class FaqEditForm(Form):
    title = StringField("Titre", validators=[required()])
    category = Select2Field("Catégorie", choices=[], validators=[required()])
    body = StringField(
        "Contenu", validators=[required()], filters=(strip,), widget=RichTextWidget()
    )



"""Blueprint (and internat service) for logs."""
from __future__ import annotations

from flask_smorest import Blueprint

from labster.security import login_required

blueprint = Blueprint("logs", __name__, template_folder="templates", url_prefix="/logs")
route = blueprint.route

__all__ = ("route",)


@blueprint.before_request
@login_required
def before_request() -> None:
    # Do nothing, just require login
    pass


@route("/")
def index():
    return ""
    # entries = LogEntry.query.all()
    #
    # return render_template("logs/index.j2", entries=entries)



from __future__ import annotations

import flask_profiler


def register_profiler(app):
    config = app.config

    if not config.get("PROFILER_ENABLE"):
        return

    password = config.get("PROFILER_PASSWORD")

    config["flask_profiler"] = {
        "enabled": True,
        "storage": {"engine": "sqlalchemy"},
        "ignore": ["^/static/.*"],
    }
    if password:
        config["flask_profiler"]["basicAuth"] = {
            "enabled": True,
            "username": "profiler",
            "password": password,
        }
    flask_profiler.init_app(app)



"""Modèle un peu simpliste de hiérarchie LDAP + info métiers spécifiques."""
from __future__ import annotations

import json
from datetime import datetime
from typing import TYPE_CHECKING, Any

import ramda as r
import structlog
from abilian.core.entities import Entity, EntityQuery
from flask_login import UserMixin
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, \
    String, Unicode, text
from sqlalchemy.orm import relationship

from labster.domain.services import dgrtt as dgrtt_service
from labster.domain.services.roles import all_roles, get_roles, has_role, \
    set_role_value

from .unites import OrgUnit
from .workflow import EN_VALIDATION

if TYPE_CHECKING:
    from .demandes import Demande
    from .roles import Role


__all__ = ("Profile", "FLUX_TENDU", "DAILY", "WEEKLY")

logger = structlog.get_logger()

FLUX_TENDU = 0
DAILY = 1
WEEKLY = 2


class RolesMixin:
    def has_role(self, role, target=None):
        # type: (Any, Any) -> bool
        """Return True if profile has given role on given target."""
        if isinstance(role, (list, tuple, set)):
            return any(self.has_role(r, target) for r in role)
        return has_role(self, role, target)  # type: ignore

    def all_roles(self):
        return all_roles(self)

    def get_roles(self, type=None):
        return get_roles(self, type)

    def set_role_value(self, role: str, value: bool, target=None):
        return set_role_value(self, role, value, target)

    def set_default_roles(self):
        self.set_role_recherche()
        self.set_role_dgrtt()

    def set_role_recherche(self):
        roles = self.ldap_dict.get("eduPersonAffiliation")
        if "faculty" in roles or "researcher" in roles:
            self.set_role_value("porteur", True)

    def set_role_dgrtt(self):
        if self.has_role("dgrtt"):
            self.set_role_value("contact dgrtt", True)
        else:
            self.set_role_value("contact dgrtt", False)


class ChercheurMixin:
    """Méthodes spécifiques au rôle "chercheur"."""

    @property
    def contacts_dgrtt(self) -> list[tuple[str, str, Profile]]:
        """Return a list of tuples: (bureau.id, bureau.nom, contact)"""
        from labster.domain.models.mapping_dgrtt import MappingDgrtt

        labo = self.laboratoire  # type: ignore
        mapping_dgrtt = MappingDgrtt.query.filter(
            MappingDgrtt.ou_recherche == labo
        ).all()

        result = []
        for bureau in dgrtt_service.BUREAUX_DGRTT:
            if bureau.id in ["AIPI 2", "Com", "Finance 2", "Finance 3", "Moyens"]:
                continue

            for m in mapping_dgrtt:
                if m.bureau_dgrtt == bureau.id:
                    result.append((bureau.id, bureau.nom, m.contact_dgrtt))
                    break
        return result


class DirectionRechercheMixin:
    """Méthodes spécifiques au rôle "direction de recherche"."""

    @property
    def stucture_dont_je_suis_le_directeur(self) -> OrgUnit | None:
        from labster.domain.models.roles import RoleType

        roles = self.get_roles(RoleType.DIRECTION)  # type: ignore
        if len(roles) > 1:
            uid = self.uid  # type: ignore
            logger.error(f"L'utilisateur {uid} a plusieurs roles de direction")
        if not roles:
            return None
        return roles[0].context

    def demandes_a_valider(self) -> list[Demande]:
        demandes = self.mes_taches()  # type: ignore
        return [d for d in demandes if d.wf_state == EN_VALIDATION.id]


class GestionnaireMixin:
    def get_membres_de_mes_structures(self) -> list[Profile]:
        from .roles import RoleType

        roles = self.get_roles(RoleType.GDL)  # type: ignore
        membres: set[Profile] = set()
        for role in roles:
            org = role.context
            membres.update(org.get_membres())

        return r.sort_by(lambda x: x.nom, list(membres))


class RechercheMixin(ChercheurMixin, DirectionRechercheMixin):
    roles: list[Role]
    laboratoire: OrgUnit

    @property
    def structure(self) -> OrgUnit:
        sous_structure = self.sous_structure
        if sous_structure:
            return sous_structure
        else:
            return self.laboratoire

    @property
    def sous_structure(self) -> OrgUnit | None:
        from .roles import RoleType

        roles = self.roles
        roles = [r for r in roles if r.type == RoleType.MEMBRE.value]
        assert len(roles) in (0, 1)
        if roles:
            return roles[0].context
        else:
            return None

    @property
    def equipe(self) -> OrgUnit | None:
        from .unites import EQUIPE

        sous_structure = self.sous_structure
        if not sous_structure:
            return None
        if sous_structure.type == EQUIPE:
            return sous_structure
        parent = sous_structure.parent
        if parent.type == EQUIPE:
            return parent
        return None

    @property
    def departement(self) -> OrgUnit | None:
        from labster.domain.models.unites import DEPARTEMENT, EQUIPE

        sous_structure = self.sous_structure
        if not sous_structure:
            return None
        if sous_structure.type == DEPARTEMENT:
            return sous_structure
        if sous_structure.type == EQUIPE:
            parent = sous_structure.parent
            if parent.type == DEPARTEMENT:
                return parent
        return None


class AgentDgrttMixin:
    @property
    def bureau_dgrtt(self):
        """Mon bureau DGRTT."""
        return dgrtt_service.get_bureau_dgrtt(self)

    @property
    def perimetre_dgrtt(self):
        """Liste des labos auprès desquels l'agent intervient."""
        return dgrtt_service.get_perimetre_dgrtt(self)


class DirectionDgrttMixin:
    pass


class WorkflowActorMixin(RolesMixin):
    # Silence the typechecker
    laboratoire: OrgUnit | None

    @property
    def stucture_dont_je_suis_le_directeur(self) -> OrgUnit | None:
        return None

    @property
    def perimetre_dgrtt(self) -> set[OrgUnit]:
        return set()


class ProfileQuery(EntityQuery):
    def get_by_uid(self, uid: str) -> Profile:
        return self.filter(Profile.uid == uid).one()


class Profile(UserMixin, AgentDgrttMixin, RechercheMixin, WorkflowActorMixin, Entity):
    __tablename__ = "profile"
    __indexable__ = False
    query_class = ProfileQuery

    #: Unique id from LDAP (ex: "jdupont"), not the same as id (surrogate key)
    uid = Column(String, nullable=False, unique=True, index=True)

    # TODO: add 'nullable=False'
    active = Column(Boolean, default=True, server_default=text("TRUE"))

    nom = Column(Unicode, nullable=False)
    prenom = Column(Unicode, nullable=False)
    email = Column(String, nullable=False)

    adresse = Column(Unicode, default="")
    telephone = Column(Unicode, default="")

    laboratoire_id = Column(Integer, ForeignKey(OrgUnit.id))
    laboratoire = relationship(
        OrgUnit, foreign_keys=[laboratoire_id], backref="membres"
    )

    #: Membre de la gouvernance ?
    gouvernance = Column(Boolean)

    #: A vraiment les droits de la gouvernance
    gouvernance_vraiment = Column(Boolean)

    #: Membre de la DGRTT
    dgrtt = Column(Boolean)
    chef_du_bureau = Column(Unicode)

    #: LDAP stuff
    fonction_structurelle_principale = Column(Unicode)

    #: More LDAP stuff
    ldap_entry = Column(String)

    #: Infos récupérées au login (CAS)
    cas_entry = Column(String)

    date_derniere_notification_vue = Column(
        DateTime, default=datetime.utcnow, nullable=False
    )
    date_last_login = Column(DateTime)

    #: FLUX_TENDU = 0
    #: DAILY = 1
    #: WEEKLY = 2
    preferences_notifications = Column(Integer, default=0)
    preferences_nb_jours_notifications = Column(Integer, default=0)

    @property
    def id(self):
        return self.uid

    @property
    def full_name(self) -> str:
        return f"{self.prenom} {self.nom}"

    @property
    def ldap_dict(self) -> dict[str, Any]:
        try:
            return json.loads(self.ldap_entry)
        except (ValueError, TypeError):
            return {}

    @property
    def cas_dict(self) -> dict[str, Any]:
        try:
            return json.loads(self.cas_entry)
        except (ValueError, TypeError):
            return {}

    def __repr__(self):
        return f"<Profile name={self.full_name} id={self.id}>"

    def __eq__(self, other):
        if not isinstance(other, Profile):
            return False
        return self.uid == other.uid

    def __hash__(self):
        return hash(self.uid)

    def nb_notifications_non_vues(self) -> int:
        from .notifications import Notification

        return (
            Notification.query.filter(Notification.user == self)
            .filter(Notification.created_at > self.date_derniere_notification_vue)
            .count()
        )

    @property
    def is_directeur(self) -> bool:
        """Vrai si le profil est vraiment le directeur de la structure."""
        return "{15079}" in self.fonction_structurelle_principale



from __future__ import annotations

from flask_sqlalchemy import SQLAlchemy
from jsonrpcserver import method

from labster.di import injector
from labster.domain2.model.profile import ProfileRepository
from labster.domain2.model.structure import StructureRepository
from labster.domain2.services.roles import Role, RoleService
from labster.rpc.queries.user import get_roles_dto_for_user
from labster.types import JSON
from labster.util import sort_by_name

structure_repo = injector.get(StructureRepository)
profile_repo = injector.get(ProfileRepository)
role_service = injector.get(RoleService)
db = injector.get(SQLAlchemy)


@method
def get_membres(structure_id: str, include_ss: bool = False) -> JSON:
    structure = structure_repo.get_by_id(structure_id)
    assert structure

    if include_ss:
        membres = role_service.get_users_with_given_role(Role.MEMBRE, structure)
    else:
        membres_affectes = role_service.get_users_with_given_role(
            Role.MEMBRE_AFFECTE, structure
        )
        membres_rattaches = role_service.get_users_with_given_role(
            Role.MEMBRE_RATTACHE, structure
        )
        membres = set(membres_affectes) | set(membres_rattaches)

    membres_tries = sort_by_name(membres)
    membres_tries = [m for m in membres_tries if m.active]

    membres_dto = [serialize_membre(m, structure) for m in membres_tries]
    return membres_dto


@method
def get_membres_rattaches_selector(structure_id: str) -> JSON:
    structure = structure_repo.get_by_id(structure_id)
    assert structure

    membres = role_service.get_users_with_given_role(Role.MEMBRE_RATTACHE, structure)
    membres = [m for m in membres if m.active]
    membres = sort_by_name(membres)

    value = [{"id": m.id, "label": m.name} for m in membres]

    all_users = list(profile_repo.get_all())
    all_users = [u for u in all_users if u.active]
    all_users.sort(key=lambda u: (u.nom, u.prenom))

    options = [{"id": u.id, "label": u.name} for u in all_users]
    return {"selector": {"value": value, "options": options}}


#
# Serialization
#
def serialize_membre(membre, structure):
    # roles_for_user = role_service.get_roles_for_user(membre)

    # structures: List[Structure] = sort_by_name(roles_for_user[Role.MEMBRE])

    r1 = role_service.has_role(membre, Role.MEMBRE_AFFECTE, structure)
    r2 = role_service.has_role(membre, Role.MEMBRE_RATTACHE, structure)
    membre_direct = r1 or r2

    # structures_dto = serialize_structures(structures)
    roles_dto = get_roles_dto_for_user(membre, base_structure=structure)

    return {
        "id": membre.id,
        "prenom": membre.prenom,
        "nom": membre.nom,
        # "structures": structures_dto,
        "membre_direct": membre_direct,
        "affecte": membre.affectation == structure.dn,
        "roles": roles_dto,
    }


# def serialize_structures(structures):
#     return [
#         {
#             "id": s.id,
#             "type": s.type_name,
#             "name": s.sigle_ou_nom,
#             "roles": [Role.MEMBRE.value],
#         }
#         for s in structures
#     ]



# Set up warnings control
from __future__ import annotations

import logging
import warnings
from typing import Callable

import sentry_sdk
import sqlalchemy.exc
from abilian.app import Application as BaseApplication
from abilian.core.celery import FlaskCelery as BaseCelery
from abilian.core.celery import FlaskLoader as CeleryBaseLoader
from abilian.i18n import babel
from flask import Flask
from flask_injector import FlaskInjector
from sentry_sdk.integrations.flask import FlaskIntegration
from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration
from werkzeug.wsgi import ClosingIterator

from .config import get_config
from .di import injector
from .extensions import redis, whoosh
from .logging import init_logging

warnings.simplefilter("ignore", category=sqlalchemy.exc.SAWarning)
# Doesn't work, despite https://github.com/Kozea/WeasyPrint/issues/312
logging.getLogger("weasyprint").setLevel(100)

__all__ = ("create_app", "Application")


# logging.basicConfig(level=logging.INFO)
# logger = structlog.get_logger()


def create_app(config: type | None = None) -> Flask:
    app = Application("labster")
    app.wsgi_app = ReverseProxied(app.wsgi_app)

    init_config(app, config)
    init_logging(app)

    with app.app_context():
        init_extensions(app)

        # This import must happen after the injector has been set up.
        from .web import init_web

        init_web(app)

        FlaskInjector(app, injector=injector)

    from .cli import register_commands

    register_commands(app)

    # Temp

    # don't use abilian-core error handlers
    app.error_handler_spec = {}

    return app


def init_config(app: Application, config: type | None):
    if config:
        # for tests
        app.setup(config)
    else:
        app.config.from_object(get_config())
        # app.config.from_mapping()
        app.setup(None)


def init_extensions(app: Flask) -> None:
    init_sentry(app)
    redis.init_app(app)
    whoosh.init_app(app)

    # Hack: we want always french
    babel.locale_selector_func = lambda: "fr"


def init_sentry(app: Flask) -> None:
    dsn = app.config.get("SENTRY_DSN")
    if not dsn:
        return

    sentry_sdk.init(
        dsn=dsn,
        integrations=[FlaskIntegration(), SqlalchemyIntegration()],
        traces_sample_rate=0.2,
        send_default_pii=True,
    )


# loader to be used by celery workers
class CeleryLoader(CeleryBaseLoader):
    flask_app_factory = "labster.app.create_app"


class CeleryApp(BaseCelery):
    loader_cls = CeleryLoader


celery = CeleryApp()


# TODO: remove this class
class Application(BaseApplication):
    # TODO: remove all of this
    # celery_app_cls = CeleryApp
    # script_manager = None
    # default_config = get_config()

    # TODO: remove
    # APP_PLUGINS = BaseApplication.APP_PLUGINS + ()
    APP_PLUGINS = ()


# From: http://flask.pocoo.org/snippets/35/
class ReverseProxied:
    """Wrap the application in this middleware and configure the front-end
    server to add these headers, to let you quietly bind this to a URL other
    than / and to an HTTP scheme that is different than what is used locally.

    In nginx:
    location /myprefix {
        proxy_pass http://192.168.0.1:5001;
        proxy_set_header Host $host;
        proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
        proxy_set_header X-Forwarded-Proto $scheme;
        proxy_set_header X-Scheme $scheme;
        proxy_set_header X-Script-Name /myprefix;
        }

    :param app: the WSGI application
    """

    def __init__(self, app: Callable) -> None:
        self.app = app

    def __call__(
        self, environ: dict[str, str], start_response: Callable
    ) -> ClosingIterator:
        script_name = environ.get("HTTP_X_SCRIPT_NAME", "")
        if script_name:
            environ["SCRIPT_NAME"] = script_name
            path_info = environ["PATH_INFO"]
            if path_info.startswith(script_name):
                environ["PATH_INFO"] = path_info[len(script_name) :]

        scheme = environ.get("HTTP_X_SCHEME", "")
        if scheme:
            environ["wsgi.url_scheme"] = scheme
        return self.app(environ, start_response)



# Listes des DN des structures pour lesquelles on doit faire ca
# on importe les membres et pas les structures avec pour règle
# que les personnes qui sont dans AD sont affectées à la structure du dessus
from __future__ import annotations

import re

SU_DN = "ou=SU,ou=Affectations,dc=chapeau,dc=fr"
DRI_DN = "ou=0107,ou=SCUN,ou=SU,ou=Affectations,dc=chapeau,dc=fr"
DRV_DNS = {
    "FL": "ou=RE,ou=SG,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "FM": "ou=M0107,ou=FACM,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "FSI": "ou=S0107,ou=UP6,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
}
DRI_ET_DRV_DNS = [DRI_DN] + list(DRV_DNS.values())

FAC_DNS = [
    "ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=FACM,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=UP6,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
]

ADMINS_DN = [
    "ou=0102,ou=988,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=0701,ou=907,ou=SCUN,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=1801,ou=918,ou=UP6,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=1901,ou=919,ou=UP6,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=2101,ou=929,ou=UP6,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=2501,ou=925,ou=UP6,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=2601,ou=926,ou=UP6,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=2701,ou=927,ou=UP6,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=3301,ou=933,ou=UP6,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=3401,ou=934,ou=UP6,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=3601,ou=936,ou=UP6,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=3701,ou=937,ou=UP6,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=3801,ou=938,ou=UP6,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=3901,ou=939,ou=UP6,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=4001,ou=940,ou=UP6,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=S0101,ou=UP6,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=67010,ou=967,ou=FACM,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=AA,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=AL,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=AN,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=AR,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=ES,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=GE,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=GR,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=HI,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=IB,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=IT,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=LA,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=LC,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=LC,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=LE,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=LF,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=MU,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=OM,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=PH,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=AD,ou=SH,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=OM,ou=HI,ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=DC,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=VPDP,ou=988,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=VPEI,ou=988,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=VPEN,ou=988,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=VPEU,ou=988,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=VPFO,ou=988,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=VPMD,ou=988,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=VPNM,ou=988,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=VPPC,ou=988,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=VPRE,ou=988,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
    "ou=VPRI,ou=988,ou=SU,ou=Affectations,dc=chapeau,dc=fr",
]

PRESIDENCE_DN = "ou=0101,ou=DUPMC,ou=SCUN,ou=SU,ou=Affectations,dc=chapeau,dc=fr"
CELSA_DN = "ou=DI,ou=CE,ou=EI,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr"


def get_parent_dn(dn: str) -> str:
    path = list(dn.split(","))
    parent_path: list[str] = path[1:]

    # Prise en compte de différents cas particuliers
    parent_dn = ",".join(parent_path)

    # Les services communs sont rattachés à l'université
    if parent_dn == "ou=SCUN,ou=SU,ou=Affectations,dc=chapeau,dc=fr":
        parent_path = parent_path[1:]

    # Les UFR de la fac de lettre sont rattachés à la fac
    if parent_dn == "ou=UF,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr":
        parent_path = parent_path[1:]

    # Les ED de la fac de lettre sont rattachés à l'institut de formation doctorale
    if parent_dn == "ou=ED,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr":
        parent_path = "ou=907,ou=SCUN,ou=SU,ou=Affectations,dc=chapeau,dc=fr".split(",")

    # Les DF de la fac de lettre sont rattachés à la fac
    if parent_dn == "ou=SG,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr":
        parent_path = parent_path[1:]

    # Cas particulier du CELSA:
    if dn == CELSA_DN:
        parent_path = parent_path[2:]

    # Les labos de la fac de lettre sont rattachés à une structure
    # "laboratoires de la fac de lettre"
    if re.match(
        "^ou=[0-9]+,ou=ED,ou=FACL,ou=SU,ou=Affectations,dc=chapeau,dc=fr$", parent_dn
    ):
        parent_path = ["ou=LABOS"] + parent_path[2:]

    return ",".join(parent_path)



from __future__ import annotations

from pytest import fixture

from labster.auth import AuthContext
from labster.di import injector
from labster.domain2.model.profile import ProfileRepository
from labster.domain2.model.structure import StructureRepository
from labster.domain2.services.roles import RoleService


@fixture(scope="session")
def structure_repo():
    return injector.get(StructureRepository)


@fixture(scope="session")
def profile_repo():
    return injector.get(ProfileRepository)


@fixture(scope="session")
def role_service():
    return injector.get(RoleService)


@fixture(scope="session")
def auth_context():
    return injector.get(AuthContext)



from __future__ import annotations

from typing import Collection

from flask_sqlalchemy import SQLAlchemy
from injector import inject
from sqlalchemy.orm import Session

from labster.domain2.model.demande import Demande, DemandeRepository
from labster.infrastructure.repositories.sqla.mappers import Mapper

# types_demande = [cls._type.value for cls in _REGISTRY.values()]


class SqlaDemandeRepository(DemandeRepository):
    session: Session

    @inject
    def __init__(self, db: SQLAlchemy, mapper: Mapper):
        self.db = db
        self.session = db.session

    def query(self):
        return self.session.query(Demande)

    def get_all(self) -> Collection[Demande]:
        return set(self.query().all())

    def put(self, demande: Demande):
        # if not demande.id:
        #     demande.id = DemandeId.new()
        self.session.add(demande)
        self.session.flush()

    def clear(self):
        self.query().delete()



"""Debug views for notifictions."""
from __future__ import annotations

from flask import render_template
from flask_sqlalchemy import SQLAlchemy

from labster.di import injector
from labster.domain2.model.profile import Profile

from . import route
from .mails import get_pending_notifications, make_notification_email, \
    make_recap_email

db = injector.get(SQLAlchemy)


#
# Weekly recaps
#
@route("/debug/recaps/")
def debug_recaps():
    users = (
        db.session.query(Profile).filter_by(active=True).order_by(Profile.login).all()
    )

    users_receiving_emails = []
    for user in users:
        email = make_recap_email(user)
        if not email:
            continue
        users_receiving_emails.append(user)

    return render_template(
        "notifications/debug/recaps.html", users=users_receiving_emails
    )


@route("/debug/recaps/<id>")
def debug_recap_for(id):
    user = Profile.query.get(id)
    email = make_recap_email(user)
    if email:
        return email
    else:
        return "Pas de mail à envoyer"


#
# Notifications (activity streams)
#
@route("/debug/streams/")
def debug_streams():
    users = (
        db.session.query(Profile).filter_by(active=True).order_by(Profile.login).all()
    )

    users_receiving_emails = []
    for user in users:
        notifications = get_pending_notifications(user)
        if not notifications:
            continue
        users_receiving_emails.append(user)

    return render_template(
        "notifications/debug/notifications.html", users=users_receiving_emails
    )


@route("/debug/streams/<id>")
def debug_stream_for(id):
    user = Profile.query.get(id)
    notifications = get_pending_notifications(user)
    email = make_notification_email(user, notifications)
    if email:
        return email
    else:
        return "Pas de mail à envoyer"



from __future__ import annotations

from enum import Enum, unique
from typing import Any, Collection

from labster.domain2.model.profile import Profile
from labster.domain2.model.structure import Structure


@unique
class Role(Enum):
    # Rôles locaux (à une structure)
    MEMBRE = "Membre"
    MEMBRE_AFFECTE = "Membre affecté"
    MEMBRE_RATTACHE = "Membre rattaché"
    MEMBRE_AFFILIE = "Membre affilié"

    RESPONSABLE = "Responsable"
    GESTIONNAIRE = "Gestionnaire Lab&Co"
    PORTEUR = "Porteur Lab&Co"
    SIGNATAIRE = "Signataire"
    ADMIN_LOCAL = "Administrateur Lab&Co local"

    # Rôles locaux (à une demande)
    CONTRIBUTEUR = "Contributeur"

    # Rôle globaux
    ADMIN_CENTRAL = "Administrateur central"
    FAQ_EDITOR = "Editeur de la FAQ"

    CONTACT = "Contact Lab&Co"


ROLE_MAP = {
    "porteur": Role.PORTEUR,
    "responsable": Role.RESPONSABLE,
    "gestionnaire": Role.GESTIONNAIRE,
    "admin_local": Role.ADMIN_LOCAL,
}


class RoleService:
    def is_empty(self):
        raise NotImplementedError

    def clear(self):
        raise NotImplementedError

    def grant_role(self, user: Profile, role: Role, context: Any = None):
        raise NotImplementedError

    def ungrant_role(self, user: Profile, role: Role, context: Any = None):
        raise NotImplementedError

    def has_role(self, user: Profile, role: Role, context: Any = None):
        raise NotImplementedError

    def get_users_with_role(self, role: Role, context: Any = None):
        raise NotImplementedError

    def get_users_with_role_on(self, context: Structure) -> dict[Role, set[Profile]]:
        raise NotImplementedError

    def get_users_with_given_role(
        self, role: Role, context: Structure
    ) -> Collection[Profile]:
        raise NotImplementedError

    def get_roles_for_user(self, user: Profile) -> dict[Role, set[Structure]]:
        raise NotImplementedError

    def update_roles(self, user: Profile):
        roles = self.get_roles_for_user(user)

        structures: set[Structure]

        structures = roles[Role.MEMBRE]
        for structure in structures:
            self.ungrant_role(user, Role.MEMBRE, structure)

        structures = roles[Role.MEMBRE_AFFILIE]
        for structure in structures:
            self.ungrant_role(user, Role.MEMBRE_AFFILIE, structure)

        structures = roles[Role.MEMBRE_AFFECTE] | roles[Role.MEMBRE_RATTACHE]

        for structure in structures:
            self.grant_role(user, Role.MEMBRE, structure)

        ancestors = set()
        for structure in structures:
            if structure:
                ancestors.update(structure.ancestors)

        for structure in ancestors:
            self.grant_role(user, Role.MEMBRE_AFFILIE, structure)
            self.grant_role(user, Role.MEMBRE, structure)



from __future__ import annotations

from typing import Any

from labster.di import injector
from labster.domain2.model.profile import ProfileRepository
from labster.domain2.services.roles import Role, RoleService
from labster.security import get_current_profile
from labster.util import url_for

profile_repository = injector.get(ProfileRepository)
role_service = injector.get(RoleService)


class Node:
    _label = ""

    def __getitem__(self, key: str) -> Any:
        return getattr(self, "_" + key)

    def get(self, key: str, default=None) -> Any:
        return getattr(self, "_" + key, default)

    def is_active(self) -> bool:
        profile = get_current_profile()
        required_roles = self.get("requires_role")
        precondition = self.get("precondition")

        if precondition:
            return precondition()

        if not required_roles:
            return True

        for role in required_roles:
            if role == "alc" and profile.has_role(Role.ADMIN_CENTRAL):
                return True

            if role == "directeur" and profile.has_role(Role.RESPONSABLE, "*"):
                return True

            if isinstance(role, Role) and profile.has_role(role):
                return True

        return False

    def __repr__(self):
        return f"<Node label={self._label}>"


class MenuItem(Node):
    _url = ""
    _endpoint = ""
    _entries: list[dict] = []

    def __init__(self, specs: dict[str, Any]):
        for k, v in specs.items():
            setattr(self, "_" + k, v)

    @property
    def url(self) -> str:
        if self._url:
            return self._url

        if self._endpoint:
            return url_for(self._endpoint)

        return ""

    @property
    def entries(self) -> list[MenuItem]:
        return [MenuItem(entry) for entry in self._entries]

    def asdict(self):
        result = {}

        for k, v in vars(self).items():
            if k.startswith("_"):
                if isinstance(v, (list, dict, int, str, type(None))):
                    result[k[1:]] = v

        result["url"] = self.url
        result["entries"] = [entry.asdict() for entry in self.entries]

        return result


class Menu(Node):
    _entries: list[dict] = []

    def __init__(self, specs: dict) -> None:
        for k, v in specs.items():
            setattr(self, "_" + k, v)
        self.entries = [MenuItem(e) for e in self._entries]

    @property
    def active_entries(self) -> list[MenuItem]:
        if not self.is_active():
            return []
        result = [e for e in self.entries if e.is_active()]
        return result

    def is_empty(self) -> bool:
        return len(self.active_entries) == 0

    def asdict(self):
        result = {}
        for k, v in vars(self).items():
            if k.startswith("_"):
                if isinstance(v, (list, dict, int, str, type(None))):
                    result[k[1:]] = v
        result["entries"] = [entry.asdict() for entry in self.active_entries]
        return result



""""""
from __future__ import annotations

import click
from abilian.core.extensions import db
from flask.cli import with_appcontext
from tqdm import tqdm

from labster.domain2.model.demande import Demande
from labster.extensions import whoosh

STOP = object()
COMMIT = object()


@click.command()
@click.option("--clear/--no-clear")
@with_appcontext
def reindex(clear: bool):
    """Reindex all content; optionally clear index before.

    :param clear: clear index content.
    """
    reindex_demandes()


def reindex_demandes():
    demandes = db.session.query(Demande).all()
    writer = whoosh.index.writer()
    for demande in tqdm(demandes, disable=None):
        whoosh.index_object(demande, writer)

    writer.commit()



from __future__ import annotations

import uuid

from flask_babel import format_date

from labster.domain.models.profiles import Profile
from labster.domain.models.util import parse_date
from labster.types import JSONDict


class Field:
    required = False
    visible = True
    hidden = False
    editable = True
    scalar = True
    note = ""
    specs: list[list[str]] = []

    def __init__(self, name, label, **kw):
        self.name = name
        self.label = label
        for k, v in kw.items():
            setattr(self, k, v)

    def to_dict(self) -> JSONDict:
        return {
            "name": self.name,
            "type": self.__class__.__name__,
            "scalar": self.scalar,
            "label": self.label,
            "required": self.required,
            "visible": self.visible,
            "hidden": self.hidden,
            "editable": self.editable,
            "note": self.note,
            "specs": self.specs,
        }

    def get_display_value(self, demande) -> str:
        """Return the name of this field for display in the form."""
        if self.name == "porteur":
            # get a Profile object, so than we get the full_name below
            # and not just its uid.
            value = demande.porteur
        else:
            value = demande.data.get(self.name, "")

        if value in (None, "None"):
            value = ""
        elif value is False:
            value = "Non"
        elif value is True:
            value = "Oui"
        elif isinstance(value, Profile):
            value = value.full_name

        return str(value)


class StringField(Field):
    pass


class DateField(Field):
    def get_display_value(self, demande):
        value = demande.data.get(self.name, "")
        date_value = parse_date(value)
        if not value:
            return ""
        return format_date(date_value, format="medium")


class IntegerField(Field):
    pass


class EmailField(Field):
    pass


class BooleanField(Field):
    pass


class Boolean2Field(Field):
    pass


class TriStateField(Field):
    pass


# def make_choices(l: List[str]):
#     return [{"value": x, "label": x} for x in l]


class Select2Field(Field):
    choices: list[str] = []

    def to_dict(self) -> JSONDict:
        d = Field.to_dict(self)
        if callable(self.choices):
            choices = self.choices()
        else:
            choices = self.choices

        d["choices"] = choices
        return d

        # if choices and isinstance(choices[0], str):
        #     d["choices"] = make_choices(choices)
        # else:
        #     d["choices"] = choices
        # return d


class MultipleSelect2Field(Field):
    choices: list[str] = []

    def to_dict(self) -> JSONDict:
        d = Field.to_dict(self)
        if callable(self.choices):
            choices = self.choices()
        else:
            choices = self.choices
        d["choices"] = choices
        return d


class TextAreaField(Field):
    pass


class HTML(Field):
    editable = False

    def __init__(self, text, name=""):
        if not name:
            name = "html-" + uuid.uuid4().hex
        super().__init__(name, text)


class ListField(Field):
    scalar = False


class ListePartenaires(ListField):
    specs = [
        ["nom_partenaire", "Nom du partenaire"],
        ["prenom_nom_contact", "Contact"],
        ["mail_contact", "Email"],
        ["telephone_contact", "Téléphone"],
    ]


class ListePartenairesContactes(ListField):
    specs = [
        ["contact", "Contact"],
        ["nom_partenaire", "Nom du partenaire"],
    ]


class ListeDivulgationsPassees(ListField):
    specs = [
        ["type_divulgation", "Type de divulgation"],
        ["titre", "Titre"],
        ["date_lieu", "Date et lieu"],
    ]


class ListeDivulgationsFutures(ListField):
    specs = [
        ["type_divulgation", "Type de divulgation"],
        ["date", "Date envisagée"],
    ]


class ListeContrats(ListField):
    specs = [
        ["contrat", "Contrat/Partenariat de recherche"],
        ["date_signature", "Date de signature du contrat"],
        ["reference", "Référence du contrat"],
    ]


class ListeMateriels(ListField):
    specs = [
        ["materiel", "Matériel"],
    ]


class ListeAutresDeclarations(ListField):
    specs = [
        ["type_protection", "Type de protection"],
        ["organisme", "Organisme ayant fait le dépôt"],
        ["exploitation", "Exploitation industrielle"],
    ]


class ListeLicencesExistantes(ListField):
    specs = [
        ["type_licence", "Type de la licence"],
        ["nom_version_licence", "Nom et version de la licence"],
    ]


class FieldSet:
    def __init__(self, name, label, fields):
        self.name = name
        self.label = label
        self.fields = fields
        self.visible = True
        self.hidden = False

    def to_dict(self) -> JSONDict:
        return {
            "name": self.name,
            "type": [self.__class__.__name__],
            "label": self.label,
            "fields": [field.name for field in self.fields],
            "visible": self.visible,
            "hidden": self.hidden,
        }

    def __repr__(self):
        return f"<FieldSet name={self.name} visible={self.visible}>"



"""Restheart related commands."""
from __future__ import annotations

from flask.cli import AppGroup, with_appcontext

import labster.bus

restheart = AppGroup("restheart")


@restheart.command("init")
@with_appcontext
def restheart_init():
    """Create needed collections in RestHeart DB."""
    labster.bus.init()


@restheart.command("sync")
@with_appcontext
def restheart_sync_all():
    """Sync (push all data) to RestHeart DB."""
    labster.bus.sync_all_objects()



from __future__ import annotations

from typing import cast

from flask_sqlalchemy import SQLAlchemy
from glom import glom
from jsonrpcserver import method
from werkzeug.exceptions import Forbidden

from labster.di import injector
from labster.domain2.model.profile import ProfileId, ProfileRepository
from labster.domain2.model.structure import StructureId, StructureRepository
from labster.domain2.services.roles import Role, RoleService
from labster.rbac import check_can_edit_roles
from labster.rpc.cache import cache
from labster.security import get_current_profile
from labster.types import JSON

structure_repo = injector.get(StructureRepository)
profile_repo = injector.get(ProfileRepository)
role_service = injector.get(RoleService)
db = injector.get(SQLAlchemy)


@method
def add_roles(structure_id: str, profile_ids: list[str], role_id: str):
    structure = structure_repo.get_by_id(StructureId(structure_id))
    check_can_edit_roles(structure)

    for profile_id in profile_ids:
        profile = profile_repo.get_by_id(ProfileId(profile_id))
        role = Role[role_id]
        role_service.grant_role(profile, role, structure)

    db.session.commit()

    cache.evict("users")
    cache.evict("structures")


@method
def delete_role(structure_id: str, profile_id: str, role_id: str):
    structure = structure_repo.get_by_id(StructureId(structure_id))
    check_can_edit_roles(structure)

    profile = profile_repo.get_by_id(ProfileId(profile_id))
    role = Role[role_id]
    role_service.ungrant_role(profile, role, structure)

    db.session.commit()

    cache.evict("users")
    cache.evict("structures")


@method
def update_roles(structure_id: str, data: dict[str, JSON]):
    structure = structure_repo.get_by_id(StructureId(structure_id))
    check_can_edit_roles(structure)

    for role_name in data:
        role = getattr(Role, role_name)

        users = role_service.get_users_with_given_role(role, structure)
        for user in users:
            role_service.ungrant_role(user, role, structure)

        values = data[role_name]
        if isinstance(values, dict):
            values = [values]
        if not values:
            continue

        for user_id in glom(values, ["id"]):
            user = profile_repo.get_by_id(ProfileId(user_id))
            role_service.grant_role(user, role, structure)

    # Cf. https://trello.com/c/bGR53cB9/33
    signataire_dto = cast(dict[str, str], data.get(Role.SIGNATAIRE.name, {}))
    if signataire_dto:
        signataire_id: str = signataire_dto["id"]
        signataire = profile_repo.get_by_id(ProfileId(signataire_id))
        role_service.grant_role(signataire, Role.RESPONSABLE, structure)

    db.session.commit()

    cache.evict("users")
    cache.evict("structures")


@method
def update_global_roles(data: dict[str, JSON]):
    user = get_current_profile()

    if not user.has_role(Role.ADMIN_CENTRAL):
        raise Forbidden

    for role_name in data:
        role = getattr(Role, role_name)

        users = role_service.get_users_with_role(role)
        for user in users:
            role_service.ungrant_role(user, role)

        values = data[role_name]
        if isinstance(values, dict):
            values = [values]
        if not values:
            continue

        for user_id in glom(values, ["id"]):
            user = profile_repo.get_by_id(ProfileId(user_id))
            role_service.grant_role(user, role)

    db.session.commit()

    cache.evict("users")



"""Formulaire PI."""
from __future__ import annotations

from typing import Any

from labster.domain.services.constants import get_constant
from labster.newforms.base import HTML, Boolean2Field, BooleanField, \
    FieldSet, Form, ListeContrats, ListeDivulgationsFutures, \
    ListeDivulgationsPassees, ListeMateriels, ListePartenairesContactes, \
    StringField, TextAreaField
#
# Fieldsets
#
from labster.newforms.common import laboratoire
from labster.newforms.pi_common import porteur_de_projet

description_technique = FieldSet(
    "description_technique",
    "Description technique",
    [
        StringField("titre", "Titre de l'invention", required=True),
        HTML("<p><em>En quoi l’invention se distingue-t-elle de l’existant ?</em></p>"),
        StringField(
            "elements_distincts", "A. Eléments distincts et avantages", required=True
        ),
        #
        BooleanField(
            "publications_proches",
            "B. Avez-vous connaissance de publications proches de l'invention?",
            required=True,
        ),
        StringField(
            "publications_proches_details",
            "Indiquez la référence",
            note="Joindre une copie des publications (avec l'onglet <b>Pièces à joindre</b>, une fois que vous aurez enregistré le formulaire).",
        ),
        #
        BooleanField(
            "recherche_bases_brevets",
            "C. Une recherche dans les bases de données brevets a-t-elle été effectuée* ? ",
            required=True,
        ),
        StringField(
            "recherche_bases_brevets_detail_1",
            "Quelle(s) base(s) de données, quels mots clés et quels documents avez-vous identifiés?",
            required=True,
        ),
        StringField(
            "recherche_bases_brevets_detail_2",
            "Parmi les documents identifiés, quel est pour vous le document le plus proche de l’invention? ",
            required=True,
        ),
        StringField(
            "recherche_bases_brevets_detail_3",
            "Précisez en quoi votre invention diffère des documents cités",
            required=True,
        ),
        #
        TextAreaField(
            "limites",
            "Quelles sont pour vous les limites de mise en œuvre de l’invention et comment peuvent-elles être surmontées?",
            required=True,
        ),
        TextAreaField(
            "applications",
            "Quelles applications potentielles de l’invention avez-vous identifiées (forces/faiblesses)?",
            required=True,
        ),
        TextAreaField(
            "prochaines_etapes",
            "Quelles sont pour vous les prochaines étapes pour les développer et les valider?",
            required=True,
        ),
        #
        BooleanField(
            "depot_anterieur",
            "Avez-vous déjà déposé une déclaration d’invention dans le domaine de l’invention ou un domaine proche?",
            required=True,
        ),
        TextAreaField(
            "depot_anterieur_detail",
            "Indiquez si cette déclaration d’invention a donné lieu à des dépôts de brevet(s)ou de logiciel(s) en mentionnant leurs références (numéro de dépôt, date de dépôt, nom du déposant, titre,…) et précisez si ce(s) brevet(s)/ logiciel(s) font l’objet d’une exploitation industrielle ou commerciale",
            required=True,
        ),
        BooleanField(
            "depot_anterieur_gere_par_upmc",
            "Cette déclaration d’invention a-t-elle été gérée par Sorbonne Université?",
            required=True,
        ),
        StringField(
            "depot_anterieur_gere_par", "Précisez qui l'a gérée", required=True
        ),
    ],
)

valorisation = FieldSet(
    "valorisation",
    "Valorisation de l'invention",
    [
        BooleanField(
            "partenaires",
            "Connaissez-vous des partenaires intéressés ou susceptibles d’être intéressés pour exploiter l’invention? ",
            required=True,
        ),
        TextAreaField(
            "precision_partenaires",
            "Indiquez le nom de ces partenaires potentiels",
            required=True,
        ),
        BooleanField(
            "partenaires_contactes",
            "Avez-vous déjà contacté certaines entreprises? ",
            required=True,
        ),
        ListePartenairesContactes(
            "liste_partenaires_contactes",
            "Indiquez le nom de ces entreprises et votre contact",
            note="Joindre la copie de l'accord de secret avec l'onglet <b>Pièces à joindre</b>, une fois que vous aurez enregistré le formulaire",
            required=True,
        ),
        BooleanField(
            "creation_entreprise",
            "Envisagez-vous de valoriser votre invention à travers la création d’entreprise?",
            required=True,
        ),
        TextAreaField(
            "precision_creation_entreprise",
            "Précisez l’avancée de ce projet",
            required=True,
        ),
    ],
)

divulgation = FieldSet(
    "divulgation",
    "Divulgation",
    [
        HTML(
            """<p>Toute divulgation des résultats valorisables avant protection affecte le caractère de « nouveauté » critère indispensable.</p>

            <p>Divulgation orale ou écrite de l’invention (par exemple : discussions avec des tiers couvertes ou non par un accord de confidentialité, publication, abstract, poster ou présentation orale en congrès, séminaire, soutenances de thèse/Master/concours/HDR sans huis clos, rapport d’activité, dépôt d’une séquence dans une banque de données…</p>"""
        ),
        #
        BooleanField(
            "divulgations_passees",
            "<u>Dans le passé</u> : l’invention a-t-elle déjà fait l’objet d’une divulgation par l’équipe de recherche ou un tiers ? y compris lors de discussions informelles avec des tiers sans accord de confidentialité* ?",
            required=True,
        ),
        ListeDivulgationsPassees(
            "liste_divulgations_passees",
            "Liste des divulgations passees",
            required=True,
        ),
        #
        BooleanField(
            "divulgations_futures",
            "<u>Dans le futur</u> : l’invention doit-elle faire prochainement l’objet d’une divulgation écrite ou orale?",
            required=True,
        ),
        ListeDivulgationsFutures(
            "liste_divulgations_futures",
            "Liste des divulgations futures",
            required=True,
        ),
    ],
)

cadre_contractuel = FieldSet(
    "cadre_contractuel",
    "Cadre contractuel et financement",
    [
        BooleanField(
            "materiels",
            "Avez-vous utilisé pour vos recherches du matériel (matériel biologique, logiciel…) obtenu auprès de sociétés ou d’autres équipes de recherche?",
            required=True,
        ),
        ListeMateriels(
            "liste_materiels",
            "Liste des materiels",
            note="Joindre une copie de l’accord de transfert de matériel (MTA) correspondant avec l’onglet Pièces à joindre, une fois que vous aurez enregistré le formulaire",
            required=True,
        ),
        #
        BooleanField(
            "contrats",
            "Les travaux de recherche à l’origine de l’invention ont-ils été conduits dans le cadre de contrat ou partenariat de recherche ?",
            note="Par exemple financement Sorbonne Université et/ou CNRS, partenariat académique, contrat de recherche avec un partenaire privé, sponsoring, CIFRE, ANR, contrat européen, financement industriel, ministère, régional, OSEO, association, fondation, chaire…",
            required=True,
        ),
        ListeContrats(
            "liste_contrats", "Liste des contrats/partenariats", required=True
        ),
    ],
)

engagement = FieldSet(
    "engagement",
    "Engagement de l'interlocuteur privilégié de l'invention",
    [
        Boolean2Field(
            "engagement",
            "En tant que porteur du projet, je reconnais accepter d’être l'interlocuteur privilégié de l'invention auprès de Sorbonne Université, et de la SATT LUTECH le cas échéant, au nom de l’ensemble des contributeurs, et confirme la validité des renseignements indiqués dans cette déclaration d'invention.",
            required=True,
        )
    ],
)


#
# The Form
#
class DemandePiInventionForm(Form):
    template = "newforms/newform.html"
    name = "pi_invention"
    label = "Déclaration d'invention / déclaration de travaux valorisables"
    conditions: Any = None

    fieldsets = [
        laboratoire,
        porteur_de_projet,
        description_technique,
        divulgation,
        cadre_contractuel,
        valorisation,
        engagement,
        # pieces_a_joindre,  # added in Form.post_init
    ]

    def post_init(self):
        """Populate form messages from json constants."""
        val = get_constant("pi_invention.pieces_a_joindre")
        pieces_a_joindre = FieldSet("pieces_a_joindre", "Pièces à joindre", [HTML(val)])
        self.fieldsets += [pieces_a_joindre]

        conditions = get_constant("pi_invention.conditions")
        self.conditions = conditions



from __future__ import annotations

from labster.domain2.model.demande import DemandeRH
from labster.domain2.model.profile import Profile, ProfileRepository
from labster.domain2.model.structure import Structure, StructureRepository
from labster.domain2.model.type_structure import DU, LA
from labster.domain2.services.roles import Role, RoleService
from labster.domain2.services.workflow.states import ABANDONNEE, \
    EN_INSTRUCTION, EN_VALIDATION, EN_VERIFICATION, TRAITEE
from labster.domain2.services.workflow.transitions import ABANDONNER, \
    ACCUSER_RECEPTION, COMMENTER, CONFIRMER_FINALISATION_DGRTT, \
    CONFIRMER_RECEVABILITE_DGRTT, DESARCHIVER, PRENDRE_LA_MAIN_DGRTT, \
    REJETER_DGRTT, REQUERIR_MODIFICATION_DGRTT, REQUERIR_MODIFICATION_DIR, \
    SOUMETTRE, VALIDER_DIR
from labster.domain2.services.workflow.workflow import LabsterWorkflow
from labster.domain.models.workflow import EN_EDITION
from labster.ldap.constants import DRI_DN
from labster.lib.workflow import State


def test_state():
    class SomeState(State):
        pass

    assert SomeState().id == "SOME_STATE"


def prepare_test(injector, db):
    # On créé in certain nombre de structures et de profils utilisateurs pour chaque tests
    class NS:
        # Create structures
        labo = Structure(dn="labo", nom="Labo", type_name=LA.name)
        dri = Structure(dn=DRI_DN, nom="DR&I", type_name=DU.name)

        # Create profiles
        joe_gdl = Profile(uid="joe_gdl", email="1", prenom="Joe", nom="GDL")
        jim_porteur = Profile(uid="jim_porteur", email="2", prenom="Jim", nom="Porteur")
        jake_directeur = Profile(
            uid="jake_directeur", email="3", prenom="Jake", nom="Directeur"
        )
        jules_dri = Profile(
            uid="jules_dgrtt", email="4", prenom="Jules", nom="Contact DR&I"
        )
        jules2_dri = Profile(
            uid="jules2_dgrtt", email="5", prenom="Jules2", nom="Contact DR&I"
        )

    x = NS()

    role_service = injector.get(RoleService)
    structure_repository = injector.get(StructureRepository)
    profile_repository = injector.get(ProfileRepository)

    structure_repository.put(x.labo)
    structure_repository.put(x.dri)

    profile_repository.put(x.jim_porteur)
    profile_repository.put(x.jake_directeur)
    profile_repository.put(x.joe_gdl)
    profile_repository.put(x.jules_dri)
    profile_repository.put(x.jules2_dri)

    role_service.grant_role(x.jake_directeur, Role.RESPONSABLE, x.labo)
    role_service.grant_role(x.jim_porteur, Role.MEMBRE, x.labo)
    role_service.grant_role(x.joe_gdl, Role.GESTIONNAIRE, x.labo)
    role_service.grant_role(x.joe_gdl, Role.GESTIONNAIRE, x.labo)
    role_service.grant_role(x.jules_dri, Role.MEMBRE, x.dri)
    role_service.grant_role(x.jules2_dri, Role.MEMBRE, x.dri)

    db.session.flush()

    assert role_service.has_role(x.jake_directeur, Role.RESPONSABLE, x.labo)
    assert role_service.has_role(x.jim_porteur, Role.MEMBRE, x.labo)
    assert role_service.has_role(x.joe_gdl, Role.GESTIONNAIRE, x.labo)
    assert role_service.has_role(x.jules_dri, Role.MEMBRE, x.dri)
    assert role_service.has_role(x.jules2_dri, Role.MEMBRE, x.dri)

    db.session.flush()

    return x


def test_abandon(request_context, injector, db):
    x = prepare_test(injector, db)
    demande = DemandeRH(gestionnaire=x.joe_gdl)

    wf = LabsterWorkflow(demande, x.joe_gdl)
    wf.start()

    assert wf.current_state() == EN_EDITION
    assert demande.active
    assert wf.possible_transitions() == [ABANDONNER]

    wf.execute_transition(ABANDONNER)
    assert wf.current_state() == ABANDONNEE
    assert not demande.active
    assert not demande.editable
    assert demande.current_owners() == set()


def test_happy_path(mocker, injector, request_context, db, db_session):
    x = prepare_test(injector, db)

    demande = DemandeRH(
        porteur=x.jim_porteur, contact_labco=x.jules_dri, structure=x.labo
    )
    mocker.patch.object(demande, "is_valid", return_value=True)

    assert x.labo.responsables == {x.jake_directeur}

    wf = LabsterWorkflow(demande, x.jim_porteur)
    wf.start()

    assert wf.current_state() == EN_EDITION
    assert demande.active
    assert demande.editable
    assert wf.possible_transitions() == [SOUMETTRE, ABANDONNER, COMMENTER]

    wf.execute_transition(SOUMETTRE)
    assert wf.current_state() == EN_VALIDATION
    assert demande.active
    assert not demande.editable
    assert demande.current_owners() == {x.jake_directeur}
    assert demande.porteur not in demande.structure.responsables
    assert wf.possible_transitions() == [ABANDONNER]

    wf.set_actor(x.jake_directeur)
    assert wf.possible_transitions() == [VALIDER_DIR, REQUERIR_MODIFICATION_DIR]

    wf.execute_transition(VALIDER_DIR)
    assert wf.current_state() == EN_VERIFICATION
    assert demande.active
    assert not demande.editable
    assert demande.current_owners() == {x.jules_dri}
    assert wf.possible_transitions() == [COMMENTER]

    wf.set_actor(x.jules_dri)
    assert wf.possible_transitions() == [
        ACCUSER_RECEPTION,
        CONFIRMER_RECEVABILITE_DGRTT,
        REQUERIR_MODIFICATION_DGRTT,
        REJETER_DGRTT,
        COMMENTER,
    ]

    wf.execute_transition(ACCUSER_RECEPTION)
    assert wf.possible_transitions() == [
        CONFIRMER_RECEVABILITE_DGRTT,
        REQUERIR_MODIFICATION_DGRTT,
        REJETER_DGRTT,
        COMMENTER,
    ]

    wf.execute_transition(CONFIRMER_RECEVABILITE_DGRTT)
    assert wf.current_state() == EN_INSTRUCTION
    assert demande.active
    assert not demande.editable
    assert demande.current_owners() == {x.jules_dri}
    assert wf.possible_transitions() == [
        CONFIRMER_FINALISATION_DGRTT,
        REQUERIR_MODIFICATION_DGRTT,
        REJETER_DGRTT,
        COMMENTER,
    ]

    wf.execute_transition(CONFIRMER_FINALISATION_DGRTT)
    assert wf.current_state() == TRAITEE
    assert not demande.active
    assert not demande.editable
    assert demande.current_owners() == set()
    assert wf.possible_transitions() == [DESARCHIVER]

    demande.id = None


def test_demande_modification(mocker, request_context, injector, db):
    x = prepare_test(injector, db)

    demande = DemandeRH(porteur=x.jim_porteur, structure=x.labo)
    mocker.patch.object(demande, "is_valid", return_value=True)

    wf = LabsterWorkflow(demande, x.joe_gdl)
    wf.start()

    wf.set_actor(x.jim_porteur)
    wf.execute_transition(SOUMETTRE)
    assert wf.current_state() == EN_VALIDATION

    wf.set_actor(x.jake_directeur)
    wf.execute_transition(REQUERIR_MODIFICATION_DIR)
    assert wf.current_state() == EN_EDITION

    wf.set_actor(x.jim_porteur)
    wf.execute_transition(SOUMETTRE)
    assert wf.current_state() == EN_VALIDATION


def test_demande_modification_dri(mocker, request_context, injector, db):
    x = prepare_test(injector, db)

    demande = DemandeRH(gestionnaire=x.joe_gdl, porteur=x.jim_porteur, structure=x.labo)
    mocker.patch.object(demande, "is_valid", return_value=True)

    wf = LabsterWorkflow(demande, x.joe_gdl)
    wf.start()

    wf.set_actor(x.jim_porteur)
    wf.execute_transition(SOUMETTRE)

    wf.set_actor(x.jake_directeur)
    wf.execute_transition(VALIDER_DIR)

    demande.contact_labco = x.jules_dri

    wf.set_actor(x.jules_dri)
    wf.execute_transition(ACCUSER_RECEPTION)
    wf.execute_transition(REQUERIR_MODIFICATION_DGRTT)

    wf.set_actor(x.jim_porteur)
    wf.execute_transition(SOUMETTRE)

    wf.set_actor(x.jules_dri)
    wf.execute_transition(CONFIRMER_RECEVABILITE_DGRTT)
    wf.execute_transition(REQUERIR_MODIFICATION_DGRTT)

    wf.set_actor(x.jim_porteur)
    wf.execute_transition(SOUMETTRE)

    wf.set_actor(x.jules_dri)
    wf.execute_transition(CONFIRMER_FINALISATION_DGRTT)

    assert not demande.active


def test_gestionnaire_peut_soumettre(mocker, request_context, injector, db):
    # si la structure permet aux gestionnaires de soumettre directement
    # le gestionnaire voit le bouton soumettre (arrive à soumettre).

    x = prepare_test(injector, db)

    demande = DemandeRH(porteur=x.jim_porteur, structure=x.labo)
    mocker.patch.object(demande, "is_valid", return_value=True)

    x.labo.permettre_soummission_directe = True

    wf = LabsterWorkflow(demande, x.joe_gdl)
    wf.start()
    wf.execute_transition(SOUMETTRE)
    assert wf.state == EN_VALIDATION


def test_prendre_la_main_dri(mocker, request_context, injector, db):
    x = prepare_test(injector, db)

    demande = DemandeRH(
        porteur=x.jim_porteur, contact_labco=x.jules_dri, structure=x.labo
    )

    mocker.patch.object(demande, "is_valid", return_value=True)
    mocker.patch.object(demande, "assigne_contact_labco")

    wf = LabsterWorkflow(demande, x.jim_porteur)
    wf.start()
    wf.execute_transition(SOUMETTRE)

    wf.set_actor(x.jake_directeur)
    wf.execute_transition(VALIDER_DIR)
    assert wf.state == EN_VERIFICATION

    # Jules2 veut prendre la main
    wf.set_actor(x.jules2_dri)
    assert wf.possible_transitions() == [
        PRENDRE_LA_MAIN_DGRTT,
        # REJETER_DGRTT,
        # COMMENTER,
    ]
    assert demande.contact_labco == x.jules_dri

    wf.execute_transition(PRENDRE_LA_MAIN_DGRTT)
    assert wf.possible_transitions() == [
        ACCUSER_RECEPTION,
        CONFIRMER_RECEVABILITE_DGRTT,
        REQUERIR_MODIFICATION_DGRTT,
        REJETER_DGRTT,
        COMMENTER,
    ]
    assert demande.contact_labco == x.jules2_dri

    wf.set_actor(x.jules_dri)
    assert wf.possible_transitions() == [
        PRENDRE_LA_MAIN_DGRTT,
        # REJETER_DGRTT,
        # COMMENTER,
    ]



from __future__ import annotations

import re
from decimal import Decimal

from labster.domain2.services.calculs_couts import sft

expected_table = """\
0	0
1	2.29
2	94.02
3	237.49
4	408.75
5	580.01
6	751.28
7	922.54
8	1093.8
9	1265.06
10	1436.32
"""


def test_sft() -> None:
    for s in expected_table.strip().split("\n"):
        t1, t2 = re.split(r"\s+", s)
        n = int(t1)
        expected = Decimal(t2)
        assert sft(600, n) == expected



from __future__ import annotations


def mes_structures(user):
    roles = user.get_roles()
    structure = None
    for role in roles:
        if role.type == "Direction":
            structure = role.context
    return [structure] + structure.descendants()



from __future__ import annotations

from typing import Any

import pytest
from sqlalchemy.orm.scoping import scoped_session

from labster.domain2.services.constants import get_constant, get_constants, \
    update_constants


def test_get_constant(db_session: scoped_session) -> None:
    val = get_constant("faq_categories")
    assert len(val) >= 1
    assert get_constant("nom_bureaux_dgrtt.MSAR").startswith("Moyens")

    with pytest.raises(KeyError):
        get_constant("foo.bar")


def test_update_constants(db_session: scoped_session) -> None:
    constants: dict[str, Any] = {}
    real_constants = get_constants()
    assert update_constants(constants) == real_constants

    missing_one = real_constants
    missing_one["faq_categories"] = []
    assert update_constants(missing_one) == real_constants



from __future__ import annotations

from labster.domain.models.demandes import AUTRE, AVENANT_CONVENTION, \
    CONVENTION, PI_INVENTION, PI_LOGICIEL, RECRUTEMENT
from labster.newforms.base import Form

from .autre import DemandeAutreForm
from .avenant_convention import DemandeAvenantConventionForm
from .convention import DemandeConventionForm
from .pi_invention import DemandePiInventionForm
from .pi_logiciel import DemandePiLogicielForm
from .rh import DemandeRHForm

TYPE_DEMANDE_TO_FORM = {
    CONVENTION: DemandeConventionForm,
    AVENANT_CONVENTION: DemandeAvenantConventionForm,
    RECRUTEMENT: DemandeRHForm,
    PI_LOGICIEL: DemandePiLogicielForm,
    PI_INVENTION: DemandePiInventionForm,
    AUTRE: DemandeAutreForm,
}


def get_form_class_for(demande):
    return TYPE_DEMANDE_TO_FORM[demande.type]


def get_form_class_by_name(name):
    for cls in TYPE_DEMANDE_TO_FORM.values():
        if cls.name == name:
            return cls
    raise KeyError(name)



from __future__ import annotations

from datetime import datetime

from flask_sqlalchemy import SQLAlchemy
from marshmallow import Schema, fields

from labster.di import injector
from labster.domain2.model.notification import Notification
from labster.rpc.registry import context_for
from labster.security import get_current_profile
from labster.types import JSONDict

db = injector.get(SQLAlchemy)


@context_for("timeline")
def timeline() -> JSONDict:
    user = get_current_profile()

    notifications = Notification.query.get_for_user(user)
    notifications_dto = NotificationSchema().dump(notifications, many=True).data

    user.date_derniere_notification_vue = datetime.utcnow()
    db.session.commit()

    ctx = {"notifications": notifications_dto}
    return ctx


class DemandeSchema(Schema):
    id = fields.String()
    nom = fields.String()
    icon_class = fields.String()


class NotificationSchema(Schema):
    created_at = fields.DateTime()
    date = fields.Function(lambda obj: str(obj.created_at.date()))
    body = fields.String()
    demande = fields.Nested(DemandeSchema)



"""Formulaire PI."""
from __future__ import annotations

from labster.newforms.base import HTML, BooleanField, DateField, FieldSet, \
    Form, IntegerField, ListePartenaires, Select2Field, StringField, \
    TextAreaField, TriStateField
from labster.newforms.common import contributeurs, laboratoire, \
    structures_concernees

contrat = FieldSet(
    "contrat",
    "",
    [
        BooleanField(
            "lien_contrat",
            "Votre demande est-elle en lien avec un contrat&nbsp;?",
            required=True,
        ),
        StringField("eotp_ou_no_dgrtt", "eOTP ou No DR&I", required=True),
    ],
)

porteur_de_projet = FieldSet(  #
    "porteur_de_projet",
    "Porteur de projet",
    [Select2Field("porteur", "Porteur de projet", choices=[], required=True)],
)

CHOICES1 = ["Coordinateur", "Partenaire", "Sous-traitant"]
CHOICES2 = [
    "ANR",
    "Commission Européenne",
    "Entreprise",
    "Financement public national",
    "FUI",
    "Institution européenne et internationale",
    "Fondation",
    "Autre",
]
CHOICES3 = [
    "Accord de collaboration",
    "Accord de consortium",
    "Contrat de prestation",
    "autre",
]

contrat_financement = FieldSet(  #
    "contrat_financement",
    "Informations sur le contrat et son financement",
    [
        Select2Field(
            "role_upmc", "Rôle de Sorbonne Université", required=True, choices=CHOICES1
        ),
        StringField("nom_ou_acronyme", "Nom ou acronyme du projet", required=True),
        TextAreaField("description_courte", "Description courte", required=True),
        Select2Field(
            "type_financeur", "Type de financeur", required=True, choices=CHOICES2
        ),
        StringField("nom_financeur", "Nom du financeur", required=True),
        BooleanField(
            "integre_entreprise",
            "Le projet intègre-t-il une entreprise&nbsp;?",
            required=True,
        ),
        BooleanField("appel_a_projets", "Appel à projets&nbsp;?", required=True),
        Select2Field("type_contrat", "Type de contrat", choices=CHOICES3),
        DateField(
            "date_depot",
            "Date limite de dépôt ou date de début souhaité du projet",
            required=True,
        ),
        IntegerField(
            "duree_previsionnelle", "Durée prévisionnelle (en mois)", required=True
        ),
    ],
)

partenaires = FieldSet(  #
    "partenaires",
    "Partenaires",
    [
        HTML(
            "<p>Vous devez saisir au moins un partenaire, sauf s'il s'agit d’un appel à projet, auquel cas la liste des partenaires est à renseigner dans les pièces jointes et non ici.</p>"
        ),
        ListePartenaires("partenaires", "Partenaires", required=False),
    ],
)

projet_scientifique = FieldSet(  #
    "projet_scientifique",
    "Projet scientifique et contributions de la structure",
    [
        HTML(
            '<p>Les documents complémentaires (canevas d’appel à projets, annexe scientifique, annexe financière…) sont à déposer dans l\'onglet "pièces à joindre".</p>'
        ),
        HTML(
            '<p>Utilisez l\'onglet "Feuille de coût" pour effectuer le calcul du coût de votre projet.</p>'
        ),
        TextAreaField(
            "role_laboratoire",
            "Description du rôle de la structure et du résultat attendu",
        ),
        HTML("<p><b>Informations complémentaires</b></p>"),
        IntegerField("montant_financement", "Montant du financement envisagé (en €)"),
        IntegerField("nombre_recrutements", "Nombre de recrutements prévisionnels"),
    ],
)

apport_upmc = FieldSet(  #
    "apports_upmc",
    "Apports scientifiques Sorbonne Université dans le projet – Résultats",
    [
        HTML(
            "<p><em>Quelles sont les résultats antérieurs mis en œuvre dans le projet par votre équipe Sorbonne Université&nbsp;?</em></p>"
        ),
        #
        TriStateField("materiel_donnees", "Matériel / données", required=True),
        BooleanField("materiel_donnees_propre_upmc", "Propre à Sorbonne Université"),
        BooleanField(
            "materiel_donnees_obtenu_precedents_contrats",
            "Obtenu lors de précédents contrats",
        ),
        BooleanField("materiel_donnees_humains", "Humains"),
        BooleanField("materiel_donnees_infectieux", "Infectieux"),
        TextAreaField("materiel_donnees_description", "Description"),
        HTML("<hr>"),
        #
        TriStateField("savoir_faire", "Savoir-faire / expertise", required=True),
        BooleanField("savoir_faire_propre_upmc", "Propre à Sorbonne Université"),
        BooleanField(
            "savoir_faire_obtenu_precedents_contrats",
            "Obtenu lors de précédents contrats",
        ),
        TextAreaField("savoir_faire_description", "Description"),
        HTML("<hr>"),
        #
        TriStateField("logiciel", "Logiciel", required=True),
        BooleanField("logiciel_propre_upmc", "Propre à Sorbonne Université"),
        BooleanField(
            "logiciel_obtenu_precedents_contrats", "Obtenu lors de précédents contrats"
        ),
        BooleanField("logiciel_libre", "Libre"),
        TextAreaField("logiciel_description", "Description"),
        HTML("<hr>"),
        #
        TriStateField("brevet", "Brevet", required=True),
        BooleanField("brevet_propre_upmc", "Propre à Sorbonne Université"),
        BooleanField(
            "brevet_obtenu_precedents_contrats", "Obtenu lors de précédents contrats"
        ),
        TextAreaField("brevet_description", "Description"),
        HTML("<hr>"),
        #
        HTML("<p><b>Types de résultats attendus</b></p>"),
        #
        BooleanField("resultats_publication", "Publication"),
        BooleanField("resultats_brevet", "Brevet"),
        BooleanField("resultats_logiciel", "Logiciel"),
        BooleanField("resultats_autre", "Autre"),
        StringField("resultats_preciser", "Préciser"),
        HTML("<hr>"),
        #
        HTML("<p><b>À qui vont appartenir les résultats&nbsp;?</b></p>"),
        #
        BooleanField("aux_tutelles_des_labos", "Aux tutelles des structures"),
        BooleanField("aux_autres_partenaires", "Aux autres partenaires"),
        BooleanField("aux_tutelles_et_partenaires", "Aux tutelles et aux partenaires"),
    ],
)

commentaires = FieldSet(  #
    "commentaires",
    "Commentaires",
    [
        HTML(
            """
<p>Vous pouvez indiquer dans ce champs des éléments complémentaires
nécessaires à l’instruction de votre demande.</p>

<p><em>Par exemple : besoin de recruter dès le démarrage du projet ;
ce nouveau contrat s’inscrit-il dans un MTA et NDA existant, etc.</em></p>"""
        ),
        TextAreaField("commentaires", "Commentaires"),
    ],
)


#
# The Form
#
class DemandeConventionForm(Form):
    template = "newforms/newform.html"
    name = "convention"
    label = "Convention de recherche"

    fieldsets = [
        contrat,
        #
        laboratoire,
        porteur_de_projet,
        structures_concernees,
        contributeurs,
        #
        contrat_financement,
        partenaires,
        projet_scientifique,
        apport_upmc,
        commentaires,
    ]



from __future__ import annotations

from flask_sqlalchemy import SQLAlchemy

from labster.bi.model import StatsLine
from labster.di import injector
from labster.domain2.model.profile import Profile
from labster.domain2.model.structure import Structure
from labster.domain2.services.roles import Role
from labster.security import get_current_profile

from .util import mes_structures

db = injector.get(SQLAlchemy)


def get_selectors():
    result = []

    types_demandes = db.session.query(StatsLine.type_demande).distinct().all()
    types_demandes.sort()
    result.append(
        {
            "name": "type_demande",
            "label": "Type de demande",
            "multiple": True,
            "options": [
                {
                    "value": x[0],
                    "text": x[0],
                }
                for x in types_demandes
            ],
        }
    )

    financeurs = (
        db.session.query(StatsLine.financeur)
        .filter(StatsLine.type_demande == "Convention de recherche")
        .filter(StatsLine.financeur != None)
        .filter(StatsLine.financeur != "")
        .order_by(StatsLine.financeur)
        .distinct()
        .all()
    )
    result.append(
        {
            "name": "financeur",
            "label": "Financeur",
            "multiple": True,
            "options": [
                {
                    "value": x[0],
                    "text": x[0],
                }
                for x in financeurs
            ],
        }
    )

    types_recrutement = (
        db.session.query(StatsLine.type_recrutement)
        .filter(StatsLine.type_demande == "Recrutement")
        .filter(StatsLine.type_recrutement != None)
        .distinct()
        .all()
    )
    result.append(
        {
            "name": "type_recrutement",
            "label": "Type de recrutement",
            "multiple": True,
            "options": [
                {
                    "value": x[0],
                    "text": x[0],
                }
                for x in types_recrutement
            ],
        }
    )

    porteur_ids = (
        db.session.query(StatsLine.porteur_id)
        .filter(StatsLine.porteur_id != None)
        .distinct()
        .all()
    )
    porteurs = [db.session.query(Profile).get(id) for id in porteur_ids]
    porteurs.sort(key=lambda x: (x.nom, x.prenom))
    result.append(
        {
            "name": "porteur_id",
            "label": "Porteur",
            # "multiple": True,
            "options": [{"value": "", "text": "-"}]
            + [
                {
                    "value": str(p.id),
                    "text": p.full_name,
                }
                for p in porteurs
            ],
        }
    )

    result.append(
        {
            "name": "structure_id",
            "label": "Structure",
            # "multiple": True,
            "options": get_structure_choices(),
        }
    )
    return result


def get_structure_choices():
    l2_ids = (
        db.session.query(StatsLine.l2).filter(StatsLine.l2 != None).distinct().all()
    )
    l3_ids = (
        db.session.query(StatsLine.l3).filter(StatsLine.l3 != None).distinct().all()
    )
    l4_ids = (
        db.session.query(StatsLine.l4).filter(StatsLine.l4 != None).distinct().all()
    )
    l5_ids = (
        db.session.query(StatsLine.l5).filter(StatsLine.l5 != None).distinct().all()
    )
    l6_ids = (
        db.session.query(StatsLine.l6).filter(StatsLine.l6 != None).distinct().all()
    )

    user = get_current_profile()
    if user.has_role(Role.RESPONSABLE, "*"):
        structures = mes_structures(user)

    else:
        ids = l2_ids + l3_ids + l4_ids + l5_ids + l6_ids
        structure_ids = [x[0] for x in ids]
        query = db.session.query(Structure)
        structures = [query.get(id) for id in structure_ids]

    to_sort = [
        (structure.path + [structure.nom], structure) for structure in structures
    ]
    to_sort.sort(key=lambda x: x[0])
    structures = [t[1] for t in to_sort]

    def make_label(structure):
        prefix = "+-" * structure.depth
        return f"{prefix}{structure.nom} ({structure.type})"

    return [{"value": "", "text": "-"}] + [
        {"value": str(s.id), "text": make_label(s)} for s in structures
    ]



from __future__ import annotations

from sqlalchemy.orm import scoped_session

from labster.di import injector
from labster.domain2.model.profile import Profile, ProfileRepository
from labster.domain2.model.structure import Structure, StructureRepository
from labster.domain2.model.type_structure import UN
from labster.domain2.services.contacts import ContactService, ContactType
from labster.domain2.services.roles import RoleService

profile_repo = injector.get(ProfileRepository)
structure_repo = injector.get(StructureRepository)
role_service = injector.get(RoleService)
contact_service = injector.get(ContactService)


def test_contact_service(db_session: scoped_session):
    # # TODO: remove
    # profile_repo.clear()
    # structure_repo.clear()
    # contact_service.clear()

    assert profile_repo.is_empty()
    assert structure_repo.is_empty()
    assert contact_service.is_empty()

    user = Profile()
    profile_repo.put(user)

    universite = Structure(nom="SU", type_name=UN.name)
    structure_repo.put(universite)

    # Actual test
    contact_service.set_contact(universite, ContactType.CDP, user)
    assert contact_service.get_contact(universite, ContactType.CDP) == user

    map = contact_service.get_mapping()
    map1 = map[universite]
    assert map1[ContactType.CDP] == user

    contact_service.delete_contact(universite, ContactType.CDP)
    assert contact_service.get_contact(universite, ContactType.CDP) == None

    map = contact_service.get_mapping()
    assert universite not in map

    # Cleanup
    profile_repo.clear()
    structure_repo.clear()
    contact_service.clear()
